# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import re
import sys
import time
from collections import defaultdict, deque
from enum import Enum
from pathlib import Path
from typing import Optional, Union, Dict, List, Set, Any, NewType, Literal
from struct import pack, unpack

from chipscopy.api import DMNodeListener
from chipscopy.api._detail.ltx import parse_ltx_files  # noqa
from chipscopy.api._detail.trace import Trace  # noqa
from chipscopy.api.containers import QueryList
from chipscopy.api.ddr.ddr import DDR
from chipscopy.api.device.client_program_log import ClientProgramLogStrategy
from chipscopy.api.device.device_spec import DeviceSpec
from chipscopy.api.hbm.hbm import HBM
from chipscopy.api.ibert import IBERT
from chipscopy.api.ila import ILA
from chipscopy.api.memory import Memory
from chipscopy.api.noc.noc import NocPerfmon
from chipscopy.api.pcie import PCIe
from chipscopy.api.sysmon import Sysmon
from chipscopy.api.vio import VIO
from chipscopy.client import ServerInfo
from chipscopy.client.axis_ila_core_client import AxisIlaCoreClient
from chipscopy.client.axis_pcie_core_client import AxisPCIeCoreClient
from chipscopy.client.axis_trace_core_client import AxisTraceClient
from chipscopy.client.axis_vio_core_client import AxisVIOCoreClient
from chipscopy.client.core import CoreParent
from chipscopy.client.ddrmc_client import DDRMCClient
from chipscopy.client.hbm_client import HBMClient
from chipscopy.client.ibert_core_client import IBERTCoreClient
from chipscopy.client.jtagdevice import JtagCable, JtagDevice
from chipscopy.client.mem import MemoryNode
from chipscopy.client.noc_perfmon_core_client import NoCPerfMonCoreClient
from chipscopy.client.sysmon_core_client import SysMonCoreClient
from chipscopy.dm import chipscope, request, Node
from chipscopy.utils.printer import printer, PercentProgressBar
from chipscopy.api.device.device_scanner import (
    scan_all_views,
)
from chipscopy.api.device.device_util import copy_node_props
from chipscopy.utils.logger import log


class FeatureNotAvailableError(Exception):
    def __init__(self):
        super().__init__("Feature not available on this architecture")


class DeviceException(Exception):
    pass


class DeviceState(Enum):
    VALID = 0
    NEEDS_REFRESH = 1
    INVALID = 2


class DeviceFamily(Enum):
    GENERIC = "generic"
    UPLUS = "ultrascaleplus"
    VERSAL = "versal"
    XVC = "xvc"

    @staticmethod
    def get_family_for_name(family_name: str) -> "DeviceFamily":
        if family_name == "versal":
            return DeviceFamily.VERSAL
        elif family_name.startswith("xvc"):
            return DeviceFamily.XVC
        elif family_name.endswith("uplus") or family_name.endswith("u"):
            return DeviceFamily.UPLUS
        else:
            return DeviceFamily.GENERIC


class Device:
    """The Device class represents a single device. It is the base class for other
    Xilinx or Generic devices. Depending on the device type, some features may be available or unavailable.

    Device has the kitchen sink of properties functions to make documentation and type hints easy.
    Child classes raise FeatureNotAvailable for functions they do not support.
    """

    _CORE_TYPE_TO_CLASS_MAP = {
        "vio": AxisVIOCoreClient,
        "ila": AxisIlaCoreClient,
        "trace": AxisTraceClient,
        # todo: remove after hw_server is updated.
        "Core": AxisTraceClient,
        "pcie": AxisPCIeCoreClient,
        "npi_nir": NoCPerfMonCoreClient,
        "ibert": IBERTCoreClient,
        "sysmon": SysMonCoreClient,
        "ddrmc_main": DDRMCClient,
        "hbm": HBMClient,
    }

    def __init__(
        self,
        *,
        hw_server: ServerInfo,
        cs_server: Optional[ServerInfo],
        device_spec: DeviceSpec,
        disable_core_scan: bool,
        disable_cache: bool,
    ):
        self.state = DeviceState.NEEDS_REFRESH
        self.cached_props = {}

        self.hw_server = hw_server
        self.cs_server = cs_server

        self._device_spec = device_spec
        self.default_target: Literal["DPC", "DAP"] = "DPC"

        self.family_name = self._device_spec.get_arch_name()
        self.device_family = DeviceFamily.get_family_for_name(self.family_name)
        self.part_name = self._device_spec.get_part_name()
        self.dna = self._device_spec.get_dna()
        self.jtag_index = self._device_spec.jtag_index

        self.disable_core_scan = disable_core_scan
        self.disable_cache = disable_cache

        self.ltx = None
        self.ltx_sources = []
        self.cores_to_scan = {}  # set during discover_and_setup_cores
        self.programming_error: Optional[str] = None  # Keep error for query after programming

        self.refresh()

        # The following filter_by becomes a dictionary with architecture, jtag_index, context, etc.
        # This is used when asking for a device by filter from the device list like:
        #    report_devices(session.devices.get(dna=".*1234"))
        # Needs improvement because QueryList filtering does not work with hierarchy
        # Should always be the last part of __init__ because of to_dict() call...
        self.filter_by = self.to_dict()

    def __str__(self) -> str:
        return f"{self.part_name}:{self.dna}:{self.context}"

    def __repr__(self) -> str:
        return self.to_json()

    def __getitem__(self, item):
        props = self.to_dict()
        if item in props:
            return props[item]
        else:
            raise AttributeError(f"No property {str(item)}")

    def _raise_if_state_invalid(self):
        if self.state == DeviceState.INVALID:
            raise DeviceException(f"Device {str(self)} is invalid - session rescan required")

    def refresh(self, *, force_update=True):
        # Refresh device properties from hardware
        self._raise_if_state_invalid()
        self.cached_props = self.scan_props(force_update=force_update)
        if not self.disable_cache:
            self.state = DeviceState.VALID

    def node_callback(self, action: DMNodeListener.NodeAction, node: Node, props: Optional[Set]):
        # Called from session when a TCF node event happens
        # TODO: TypeError: 'ParodyLogger' object is not callable
        # log.client.trace(f"node_callback: action={action}, node={node.ctx}")

        if not node or action == DMNodeListener.NodeAction.NODE_ADDED:
            # Maybe in the future we will track if a child node was added to anything in the
            #     Device tree... For now just ignore.
            return

        # Top level device contexts that can represent the device - track for device removal
        root_device_contexts = {self._device_spec.device_ctx, self._device_spec.jtag_device_ctx}

        # Expanded set of device contexts that are important to track for node changes
        all_device_contexts = {
            self._device_spec.device_ctx,
            self._device_spec.jtag_device_ctx,
            self._device_spec.debugcore_dap_ctx,
            self._device_spec.debugcore_dpc_ctx,
            self._device_spec.memory_dap_ctx,
            self._device_spec.memory_dpc_ctx,
            self._device_spec.chipscope_dap_ctx,
            self._device_spec.chipscope_dpc_ctx,
        }
        if action == DMNodeListener.NodeAction.NODE_REMOVED and node.ctx in root_device_contexts:
            # Removed a node that is the anchor for this device... Invalidate the device so any subsequent
            #     user API calls return an invalid device message

            # TODO: TypeError: 'ParodyLogger' object is not callable
            # log.client.warn(f"DEVICE {str(self)} marked invalid -- rescan required!")
            self.state = DeviceState.INVALID
            device_node = self._device_spec.get_device_node()
            if device_node:
                # Remove any cached information for the invalid device
                device_node.device_wrapper = None
        elif action == DMNodeListener.NodeAction.NODE_CHANGED and node.ctx in all_device_contexts:
            # One of our watched nodes changed - need to re-scan properties from hardware on next call

            # TODO: TypeError: 'ParodyLogger' object is not callable
            # log.client.debug(
            #     f"NODE_CHANGED on device {str(self)}, Node: {node.ctx} - setting NEEDS_REFRESH"
            # )
            self.state = DeviceState.NEEDS_REFRESH

    def scan_props(self, *, force_update=True) -> Dict[str, Any]:
        """
        Scan device properties from hardware.

        Returns: Dict of properties

        """
        self._raise_if_state_invalid()
        props = {}
        if self.hw_server and self.jtag_node:
            node = self.jtag_node
            props["jtag"] = copy_node_props(node, node.props, force_update=force_update)
        props["is_valid"] = self.is_valid
        props["context"] = self.context
        props["is_programmed"] = self.scan_is_programmed(force_update=force_update)
        props["is_programmable"] = self.scan_is_programmable()
        props["part"] = self.part_name
        props["family"] = self.family_name
        props["dna"] = self.dna
        props["jtag_index"] = self.jtag_index
        props["cable_context"] = self._device_spec.jtag_cable_ctx
        props["jtag_context"] = self._device_spec.jtag_device_ctx
        props["cable_name"] = self._device_spec.get_jtag_cable_name()
        props = dict(sorted(props.items()))  # noqa
        return props

    def to_dict(self) -> Dict[str, Any]:
        """Returns a dictionary representation of the device data"""
        self._raise_if_state_invalid()
        if self.state == DeviceState.NEEDS_REFRESH:
            self.refresh()
        return self.cached_props

    def to_json(self) -> str:
        """Returns a json representation of the device data"""
        raw_json = json.dumps(self.to_dict(), indent=4, default=lambda o: str(o))
        return raw_json

    # NODE / LOW LEVEL

    @property
    def context(self) -> str:
        """Returns the unique context for this device. Normally this is the
        jtag node context. If this device is not on a jtag chain, the top
        level device context will be a non-jtag node.
        """
        return self._device_spec.device_ctx

    def scan_is_programmed(self, *, force_update=True) -> bool:
        """Returns True if this Device is programmed, False otherwise."""
        # TODO: Queries hardware for jtag program status. This could be slow -
        #       is there a way to collect this once, then track by subsequent events pushed from
        #       the hw_server if we detect DONE went low?
        self._raise_if_state_invalid()
        is_programmed_ = False
        try:
            node = self.jtag_node
            is_programmable = node.props.get("is_programmable")
            if is_programmable:
                regs = node.props.get("regs")
                if regs:
                    jtag_status = node.props.get("regs").get("jtag_status")
                    config_status = node.props.get("regs").get("config_status")
                    if jtag_status:
                        if force_update:
                            node.update_regs(reg_names=("jtag_status",), force=True, done=None)
                        done_reg = jtag_status.fields[0]["DONE"]
                        is_programmed_ = done_reg["value"] == 1
                    elif config_status:
                        if force_update:
                            node.update_regs(reg_names=("config_status",), force=True, done=None)
                        done_reg = config_status.fields["DONE_PIN"]
                        is_programmed_ = done_reg["value"] == 1
        except Exception:  # noqa
            is_programmed_ = False
        return is_programmed_

    @property
    def is_programmed(self) -> bool:
        """Returns True if this Device is programmed, False otherwise."""
        d = self.to_dict()
        return d.get("is_programmed", False)

    def scan_is_programmable(self) -> bool:
        """Returns True if this Device is programmable, False otherwise"""
        is_programmable_ = False
        node = self.jtag_node
        if node:
            return node.props.get("is_programmable", False)

    @property
    def is_programmable(self) -> bool:
        """Returns True if this Device is programmable, False otherwise"""
        d = self.to_dict()
        return d.get("is_programmable", False)

    @property
    def is_valid(self) -> bool:
        """Returns True if this Node is valid, False otherwise."""
        return self._device_spec.is_valid

    @property
    def jtag_cable_node(self) -> Optional[JtagCable]:
        """Low level TCF node access - For advanced use"""
        return self._device_spec.get_jtag_cable_node()

    @property
    def jtag_node(self) -> Optional[JtagDevice]:
        """Low level TCF node access - For advanced use"""
        return self._device_spec.get_jtag_node()

    @property
    def debugcore_node(self) -> Optional[CoreParent]:
        """Low level TCF node access - For advanced use"""
        return self._device_spec.get_debugcore_node(target=None, default_target=self.default_target)

    @property
    def chipscope_node(self) -> Optional[CoreParent]:
        """Low level TCF node access - For advanced use"""
        return self._device_spec.get_chipscope_node(target=None, default_target=self.default_target)

    # DEBUG CORES

    @property
    def ibert_cores(self) -> QueryList[IBERT]:
        """Returns:
        list of detected IBERT cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(IBERTCoreClient)

    @property
    def vio_cores(self) -> QueryList[VIO]:
        """Returns:
        list of detected VIO cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(AxisVIOCoreClient)

    @property
    def ila_cores(self) -> QueryList[ILA]:
        """Returns:
        list of detected ILA cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(AxisIlaCoreClient)

    @property
    def trace_cores(self) -> QueryList[Trace]:
        """Returns:
        list of detected Trace cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(AxisTraceClient)

    @property
    def pcie_cores(self) -> QueryList[PCIe]:
        """Returns:
        list of detected PCIe cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(AxisPCIeCoreClient)

    @property
    def noc_core(self) -> QueryList[NocPerfmon]:
        """Returns:
        List with NOC roots for the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(NoCPerfMonCoreClient)

    @property
    def ddrs(self) -> QueryList[DDR]:
        """Returns:
        list of detected DDRMC cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(DDRMCClient)

    @property
    def hbms(self) -> QueryList[HBM]:
        """Returns:
        list of detected HBM cores in the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(HBMClient)

    @property
    def sysmon_root(self) -> QueryList[Sysmon]:
        """Returns:
        List with one SysMon core for the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        return self._get_debug_core_wrappers(SysMonCoreClient)

    @staticmethod
    def _set_client_wrapper(node, api_client_wrapper):
        # Tacks a wrapper onto a node for tracking. Always call this
        # instead of setting directly on node - in case we change the
        # client wrapper tracking process.
        node.api_client_wrapper = api_client_wrapper

    @staticmethod
    def _get_client_wrapper(node):
        # Gets the tacked on wrapper from a node.
        return node.api_client_wrapper

    def _create_debugcore_wrapper(self, node):
        # Factory to build debug_core_wrappers
        # Instantiate debug core wrappers for a known core type. The wrapper gets attached to a
        # node for safe keeping. Any time we need to grab a debug core wrapper (like ILA)
        # it is available on the node using
        #     wrapper = self._get_client_wrapper(node)
        if node.type == "ila":
            debug_core_wrapper = ILA(node, self, ltx=self.ltx)
        elif node.type == "vio":
            debug_core_wrapper = VIO(node, ltx=self.ltx)
        elif node.type == "trace":
            debug_core_wrapper = Trace(node)
        # todo: remove after hw_server is updated.
        elif node.type == "Core":
            debug_core_wrapper = Trace(node)
        elif node.type == "pcie":
            debug_core_wrapper = PCIe(node)
        elif node.type == "npi_nir":
            debug_core_wrapper = NocPerfmon(node)
        elif node.type == "ibert":
            debug_core_wrapper = IBERT(node)
        elif node.type == "sysmon":
            debug_core_wrapper = Sysmon(node)
        elif node.type == "ddrmc_main":
            debug_core_wrapper = DDR(node)
        elif node.type == "hbm":
            debug_core_wrapper = HBM(node)
        else:
            # TODO: extend this factory to other core types with a plugin model for additional types
            debug_core_wrapper = None
        return debug_core_wrapper

    def _find_all_debugcore_nodes(self) -> Dict:
        # Traverse the chipscope view for debug cores in this device.
        # Returns a dict mapping core_types to a list of found nodes.
        # Notes are returned in upgraded form.
        #   Example return:
        #         {
        #           AxisIlaCoreClient: [ila_node_1, ila_node_2, ...],
        #           AxisVIOCoreClient: [vio_node_1, vio_node_2, ...],
        #           ...
        #         }
        # The first time calling this function is slow because nodes have
        # not been upgraded yet. Slow is around 1 second. Subsequent calls are
        # faster, around 1 ms.
        #
        # Best to avoid calling this function in an inner loop.

        cs_view = self.cs_server.get_view(chipscope)
        found_cores = defaultdict(list)

        def check_and_upgrade_node(node):
            if self.cores_to_scan.get(cs_node.type, True):
                core_class_type = Device._CORE_TYPE_TO_CLASS_MAP.get(node.type, None)
                if core_class_type and core_class_type.is_compatible(node):
                    upgraded_node = cs_view.get_node(node.ctx, core_class_type)
                    found_cores[core_class_type].append(upgraded_node)

        # Using the self.chipscope_node restricts nodes to this device only
        for cs_node in cs_view.get_children(self.chipscope_node):
            if cs_node.type == "debug_hub":
                debug_hub_children = cs_view.get_children(cs_node)
                for child in debug_hub_children:
                    check_and_upgrade_node(child)
            else:
                check_and_upgrade_node(cs_node)
        return found_cores

    def _get_debug_core_wrappers(self, core_type) -> QueryList:
        # Returns a QueryList of the debug_core_wrappers matching the given core_type
        #   core_type is a class from CORE_TYPE_TO_CLASS_MAP
        #   returned wrapper_type is ILA, VIO, etc. wrapping the node
        #
        # debug_core_wrappers are initialized if they don't yet exist for the
        # node. After initialization, they are hung on the node for future reference.
        # This is convenient because their lifecycle is tied to the node.
        found_cores = QueryList()
        all_debug_core_nodes = self._find_all_debugcore_nodes()
        for node in all_debug_core_nodes.get(core_type, []):
            debug_core_wrapper = Device._get_client_wrapper(node)
            if debug_core_wrapper is None:
                # Lazy debugcore initialization here.
                # Initialization happens the first time we access the debug core.
                debug_core_wrapper = self._create_debugcore_wrapper(node)
            self._set_client_wrapper(node, debug_core_wrapper)
            # Now the node should have a proper debug_core_wrapper attached.
            if Device._get_client_wrapper(node):
                found_cores.append(Device._get_client_wrapper(node))
        return found_cores

    def discover_and_setup_cores(
        self,
        *,
        hub_address_list: Optional[List[int]] = None,
        ltx_file: Union[Path, str] = None,
        **kwargs,
    ):
        """Scan device for debug cores. This may take some time depending on
        what gets scanned.

        Args:
            hub_address_list: List of debug hub addresses to scan
            ltx_file: LTX file (which contains debug hub addresses)

        Keyword Arguments:
            ila_scan: True=Scan Device for ILAs
            vio_scan: True=Scan Device for VIOs
            ibert_scan: True=Scan Device for IBERTs
            pcie_scan: True=Scan Device for PCIEs
            noc_scan: True=Scan Device for NOC
            ddr_scan: True=Scan Device for DDRs
            hbm_scan: True=Scan Device for HBMs
            sysmon_scan: True=Scan Device for System Monitor
        """
        # Selectively disable scanning of cores depending on what comes in
        # This is second priority to the disable_core_scan in __init__.
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        self.cores_to_scan = {}
        if self.disable_core_scan:
            for core_name in Device._CORE_TYPE_TO_CLASS_MAP.keys():
                self.cores_to_scan[core_name] = False
        else:
            self.cores_to_scan = {
                "ila": kwargs.get("ila_scan", True),
                "vio": kwargs.get("vio_scan", True),
                "npi_nir": kwargs.get("noc_scan", False),
                "ddrmc_main": kwargs.get("ddr_scan", True),
                "hbm": kwargs.get("hbm_scan", True),
                "pcie": kwargs.get("pcie_scan", True),
                "ibert": kwargs.get("ibert_scan", True),
                "sysmon": kwargs.get("sysmon_scan", False),
            }
        if not self.cs_server:
            raise RuntimeError("cs_server is not connected")

        self.ltx = None

        # Make sure debug_hub_addrs is a list of hub addresses.
        # It may come as an int or a list or None
        if isinstance(hub_address_list, int):
            hub_address_list = [hub_address_list]
        if not hub_address_list:
            hub_address_list = []

        if ltx_file:
            self.ltx, self.ltx_sources, _ = parse_ltx_files(ltx_file)  # noqa
            hub_address_list.extend(self.ltx.get_debug_hub_addresses())
            # Remove any duplicate hub addresses in list, in case same one
            # was passed to function as well as in ltx
            # Below is a quick python trick to uniquify a list.
            hub_address_list = list(dict.fromkeys(hub_address_list))

        # Set up all debug cores in hw_server for the given hub addresses
        # Do this by default unless the user has explicitly told us not to with
        # "disable_core_scan" as an argument.
        if not self.disable_core_scan or self.device_family == DeviceFamily.UPLUS:
            if not self.cs_server:
                raise RuntimeError("No chipscope server connection. Could not get chipscope view")
            cs_node = self.chipscope_node
            if cs_node:
                cs_node.setup_cores(debug_hub_addrs=hub_address_list)
            else:
                # print(self._device_spec.to_json())
                raise RuntimeError("chipscope_node not available")

    # MEMORY

    @property
    def memory_target_names(self) -> List[str]:
        """Returns:
        list of supported memory targets"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        valid_memory_targets = (
            r"(Versal .*)|(DPC)|(APU)|(RPU)|(PPU)|(PSM)|(Cortex.*)|(MicroBlaze.*)|(RISC-V.*)"
        )
        memory_view = self.hw_server.get_view("memory")
        memory_node_list = []
        mem_dpc_node = self._device_spec.get_memory_node(target="DPC")
        if mem_dpc_node:
            memory_node_list.append(mem_dpc_node)
        mem_dap_node = self._device_spec.get_memory_node(target="DAP")
        if mem_dap_node:
            memory_node_list.append(mem_dap_node)

        nodes_to_travel = deque(memory_node_list)
        memory_node_list = []
        while nodes_to_travel:
            node = nodes_to_travel.popleft()
            if MemoryNode.is_compatible(node):
                memory_node = memory_view.get_node(node.ctx, MemoryNode)
                memory_node_list.append(memory_node)
            nodes_to_travel.extendleft(memory_view.get_children(node))

        # Now the memory_node_list has all memory nodes.
        # Make the list of what we found and return it
        memory_target_names = []
        for node in memory_node_list:
            memory_target_name = node.props.get("Name")
            if memory_target_name and re.search(valid_memory_targets, memory_target_name):
                memory_target_names.append(memory_target_name)
        return memory_target_names

    @property
    def memory(self) -> QueryList[Memory]:
        """Returns:
        the memory access for the device"""
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        target_nodes = self._device_spec.get_all_memory_nodes()
        node_list = QueryList()
        for node in target_nodes:
            node_list.append(node)
        return node_list

    def memory_read(self, address: int, num: int = 1, *, size="w", target=None) -> List[int]:
        """Read num values from given memory address.  This method is slow because it locates the memory context
        in the target tree every time. If you want to execute a large number of memory operations, grab a memory
        context and do all the operations in batch.
        Note: This method should not be used in inner loops. It is not the fastest because it looks up the memory
        target every time. Inner loops should just call the find_memory_target once.
        """
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        node = self._device_spec.search_memory_node_deep(target, default_target=self.default_target)
        if not node:
            raise (ValueError(f"Could not find memory node for target {target}"))
        retval = node.memory_read(address=address, num=num, size=size)
        return retval

    def memory_write(self, address: int, values: List[int], *, size="w", target=None):
        """Write list of values to given memory address. This method is slow because it locates the memory context
        in the target tree every time. If you want to execute a large number of memory operations, grab a memory
        context and do the operations in batch.
        Note: This method should not be used in inner loops. It is not the fastest because it looks up the memory
        target every time. Inner loops should just call the find_memory_target once.
        """
        self._raise_if_family_not(DeviceFamily.VERSAL)
        node = self._device_spec.search_memory_node_deep(target, default_target=self.default_target)
        if not node:
            raise (ValueError(f"Could not find memory node for target {target}"))
        node.memory_write(address=address, values=values, size=size)

    def _dpc_read_bytes(
        self,
        address: int,
        data: bytearray,
        *,
        offset: int = 0,
        byte_size: int = None,
        show_progress_bar: bool = True,
    ):
        """
        Reads byte data from memory, using the DPC.

        Args:
            address (int): Start address
            data (bytearray): result buffer which gets filled with read data.
            offset (int): start index in *data* to start write to.
            byte_size (int): Number of bytes to read. Default is the size of the *data* buffer.
            show_progress_bar (bool): Show if True.
        """
        self._raise_if_state_invalid()
        node = self.debugcore_node
        if not show_progress_bar:
            node.read_bytes(address, data, offset, byte_size)
            return

        # With progress bar
        transfer_byte_count = byte_size if byte_size else len(data)
        desc = (
            f"Transferring {transfer_byte_count/(1024 * 1024):0.1f} MB "
            f"[bold](Read {hex(address)})[/]"
        )

        bar = PercentProgressBar()
        bar.add_task(description=desc, status=PercentProgressBar.Status.STARTING)

        try:
            node.read_bytes(
                address,
                data,
                offset,
                byte_size,
                progress=lambda progress: bar.update(
                    completed=int(progress * 100), status=PercentProgressBar.Status.IN_PROGRESS
                ),
            )
        except Exception as ex:
            bar.update(status=PercentProgressBar.Status.ABORTED)
            raise ex

        bar.update(status=PercentProgressBar.Status.DONE)

    def _dpc_write_bytes(
        self,
        address: int,
        data: bytearray,
        *,
        offset: int = 0,
        byte_size: int = None,
        show_progress_bar: bool = True,
    ):
        """
        Write byte data from memory, using the DPC.

        Args:
            address (int): Start address
            data (bytearray): data to write.
            offset (int): start index in *data* to start write from.
            byte_size (int): Number of bytes to write. Default is the size of the *data* buffer.
            show_progress_bar (bool): Show if True.
        """
        self._raise_if_state_invalid()
        node = self.debugcore_node
        if not show_progress_bar:
            node.write_bytes(address, data, offset, byte_size)
            return

        # With progress bar
        transfer_byte_count = byte_size if byte_size else len(data)
        desc = (
            f"Transferring {transfer_byte_count/(1024 * 1024):0.1f} MB "
            f"[bold](Write {hex(address)})[/]"
        )

        bar = PercentProgressBar()
        bar.add_task(description=desc, status=PercentProgressBar.Status.STARTING)

        try:
            node.write_bytes(
                address,
                data,
                offset,
                byte_size,
                progress=lambda progress: bar.update(
                    completed=int(progress * 100), status=PercentProgressBar.Status.IN_PROGRESS
                ),
            )
        except Exception as ex:
            bar.update(status=PercentProgressBar.Status.ABORTED)
            raise ex

        bar.update(status=PercentProgressBar.Status.DONE)

    # PROGRAMMING
    def get_plm_log(self, memory_target="Versal.+", slr_index: int = 0) -> str:
        """
        Returns: Raw log read from hardware

        Args:
            memory_target: Optional name of memory target such as APU, RPU, DPC
            slr_index: Optional SLR index for multi-SLR devices
        """
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        try:
            mem = self.memory.get(name=memory_target)
        except ValueError:
            raise ValueError(f"Memory target {memory_target} not available")
        return mem.plm_log(slr_index=slr_index)

    def get_program_log(self, memory_target="Versal.+") -> str:
        """
        Returns: Programming log read from hardware (None=Use default transport)

        Args:
            memory_target: Optional name of memory target such as APU, RPU, DPC
            slr_index: Optional SLR index for multi-SLR devices
        """
        self._raise_if_family_not(DeviceFamily.VERSAL)
        self._raise_if_state_invalid()
        try:
            mem = self.memory.get(name=memory_target)
        except ValueError:
            raise ValueError(f"Memory target {memory_target} not available")
        report_generator = ClientProgramLogStrategy(
            regs=self.jtag_node.props.get("regs", {}), plm_log_fn=lambda x: mem.plm_log(slr_index=x)
        )
        return report_generator.generate_program_log()

    def write_program_log(self, outfile: Union[str, Path], *, memory_target="Versal.+"):
        """
        Write an enhanced program log to a file.
        """
        if isinstance(outfile, str):
            outfile = Path(outfile)
        try:
            report_str = self.get_program_log(memory_target=memory_target)
            if outfile is None:
                sys.stdout.write(report_str)
            else:
                with outfile.open("w") as file:
                    file.write(report_str)
        except Exception as e:
            raise IOError(f"Failed to write report to {outfile}: {str(e)}")

    def reset(self):
        """Reset the device to a non-programmed state."""
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        jtag_programming_node = self.jtag_node
        assert jtag_programming_node is not None
        self.state = DeviceState.NEEDS_REFRESH
        jtag_programming_node.future().reset()
        time.sleep(0.5)
        self.refresh()

    def program(
        self,
        programming_file: Union[str, Path],
        *,
        skip_reset: bool = False,
        delay_after_program: int = 0,
        show_progress_bar: bool = True,
        progress: request.ProgressFutureCallback = None,
        done: request.DoneFutureCallback = None,
    ):
        """Program the device with a given programming file (bit or pdi).

        Args:
            programming_file: PDI file path
            skip_reset: False = Do not reset device prior to program
            delay_after_program: Seconds to delay at end of programming (default=0)
            show_progress_bar: False if the progress bar doesn't need to be shown
            progress: Optional progress callback
            done: Optional async future callback
        """
        self.programming_error = None
        self._raise_if_family_not(DeviceFamily.VERSAL, DeviceFamily.UPLUS)
        self._raise_if_state_invalid()
        program_future = request.CsFutureSync(done=done, progress=progress)
        self.state = DeviceState.NEEDS_REFRESH

        if isinstance(programming_file, str):
            programming_file = Path(programming_file)

        if not programming_file.exists():
            raise FileNotFoundError(
                f"Programming file: {str(programming_file.resolve())} doesn't exists!"
            )

        printer(f"Programming device with: {str(programming_file.resolve())}\n", level="info")

        if show_progress_bar:
            progress_ = PercentProgressBar()
            progress_.add_task(
                description="Device program progress",
                status=PercentProgressBar.Status.STARTING,
                visible=show_progress_bar,
            )

        def progress_update(future):
            if program_future:
                program_future.set_progress(future.progress)

            if show_progress_bar:
                progress_.update(
                    completed=future.progress * 100, status=PercentProgressBar.Status.IN_PROGRESS
                )

        def pre_done_programming(future):
            # refresh jtag node properties after program (1190699)
            if future.error:
                self.programming_error = future.error
            jtag_programming_node.future().update_regs(
                reg_names=[], force=True, done=done_programming
            )

        def done_programming(future):
            if not self.programming_error:
                self.programming_error = future.error

            if show_progress_bar:
                if self.programming_error is not None:
                    progress_.update(status=PercentProgressBar.Status.ABORTED)
                else:
                    for i in range(delay_after_program):
                        # This pushes end-of-config node events to the listeners
                        # They seem to happen within about 2-3 seconds - need a more
                        # reliable way to wait for them to finish.
                        # If we don't wait, the DPC node may not be ready for use.
                        time.sleep(0.5)
                        if self.cs_server:
                            self.cs_server.get_view("chipscope").run_events()
                        time.sleep(0.5)
                    # refresh device jtag node properties after program (1190699)
                    self.refresh(force_update=False)
                    progress_.update(completed=100, status=PercentProgressBar.Status.DONE)

            if self.programming_error is not None:
                program_future.set_exception(self.programming_error)
            else:
                program_future.set_result(None)

        def finalize_program():
            if self.programming_error is not None:
                raise RuntimeError(self.programming_error)

        jtag_programming_node = self.jtag_node
        assert jtag_programming_node is not None
        if not skip_reset:
            jtag_programming_node.future().reset()

        jtag_programming_node.future(
            progress=progress_update, done=pre_done_programming, final=finalize_program
        ).config(str(programming_file.resolve()))

        return program_future if done else program_future.result

    def _raise_if_family_not(self, *args):
        if self.device_family == DeviceFamily.XVC:
            # special partial setup case for xvc sw testing - always ok
            return
        if self.device_family not in args:
            raise FeatureNotAvailableError()


# Legacy support - to not break old code that may have referenced the classes
# that were removed (and now all implemented by Device)

VersalDevice = NewType("VersalDevice", Device)
UltrascaleDevice = NewType("UltrascaleDevice", Device)
GenericDevice = NewType("GenericDevice", Device)


def discover_devices(
    hw_server: ServerInfo,
    cs_server: ServerInfo = None,
    disable_core_scan: bool = False,
    cable_ctx: Optional[str] = None,
    disable_cache: bool = False,
) -> QueryList[Device]:
    log.client.debug(
        f"discover_devices: hw_server={hw_server}, cs_server={cs_server}, disable_core_scan={disable_core_scan}, cable_ctx={cable_ctx}",
        disable_cache={disable_cache},
    )
    include_dna = True
    devices: QueryList[Device] = QueryList()
    idx = 0
    for key, device_record_list in scan_all_views(hw_server, cs_server, include_dna).items():
        device_spec = DeviceSpec.create_from_device_records(
            hw_server, cs_server, device_record_list
        )
        if not device_spec.is_valid:
            # This can happen when a chipscope server is serving multiple sessions.
            # The cs_server reports back connection information for the
            # other sessions which should be ignored.
            log.client.debug(f"discover_devices: No device_node for dna key: {key} --> ignored.\n")
            continue

        if cable_ctx is None or device_spec.jtag_cable_ctx == cable_ctx:
            device_node = device_spec.get_device_node()
            assert device_node is not None
            if device_node.device_wrapper is None:
                device_node.device_wrapper = Device(
                    hw_server=hw_server,
                    cs_server=cs_server,
                    device_spec=device_spec,
                    disable_core_scan=disable_core_scan,
                    disable_cache=disable_cache,
                )
                log.client.debug(
                    f"discover_devices: created new device: {device_node.device_wrapper}"
                )
            else:
                log.client.debug(
                    f"discover_devices: reusing device wrapper: {device_node.device_wrapper}"
                )
            devices.append(device_node.device_wrapper)
            log.client.info(f"discover_devices: {idx}: {device_node.device_wrapper}")
            idx += 1
    return devices



# Generated from ILATsm.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def serializedATN():
    return [
        4,
        1,
        31,
        149,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        2,
        8,
        7,
        8,
        2,
        9,
        7,
        9,
        2,
        10,
        7,
        10,
        1,
        0,
        4,
        0,
        24,
        8,
        0,
        11,
        0,
        12,
        0,
        25,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        36,
        8,
        1,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        1,
        2,
        3,
        2,
        74,
        8,
        2,
        1,
        3,
        4,
        3,
        77,
        8,
        3,
        11,
        3,
        12,
        3,
        78,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        3,
        5,
        92,
        8,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        5,
        5,
        97,
        8,
        5,
        10,
        5,
        12,
        5,
        100,
        9,
        5,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        1,
        6,
        3,
        6,
        109,
        8,
        6,
        1,
        7,
        1,
        7,
        1,
        7,
        1,
        7,
        1,
        7,
        1,
        7,
        1,
        7,
        3,
        7,
        118,
        8,
        7,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        1,
        8,
        3,
        8,
        137,
        8,
        8,
        1,
        9,
        1,
        9,
        1,
        9,
        3,
        9,
        142,
        8,
        9,
        1,
        10,
        1,
        10,
        1,
        10,
        3,
        10,
        147,
        8,
        10,
        1,
        10,
        0,
        1,
        10,
        11,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        16,
        18,
        20,
        0,
        0,
        157,
        0,
        23,
        1,
        0,
        0,
        0,
        2,
        35,
        1,
        0,
        0,
        0,
        4,
        73,
        1,
        0,
        0,
        0,
        6,
        76,
        1,
        0,
        0,
        0,
        8,
        80,
        1,
        0,
        0,
        0,
        10,
        91,
        1,
        0,
        0,
        0,
        12,
        108,
        1,
        0,
        0,
        0,
        14,
        117,
        1,
        0,
        0,
        0,
        16,
        136,
        1,
        0,
        0,
        0,
        18,
        141,
        1,
        0,
        0,
        0,
        20,
        146,
        1,
        0,
        0,
        0,
        22,
        24,
        3,
        2,
        1,
        0,
        23,
        22,
        1,
        0,
        0,
        0,
        24,
        25,
        1,
        0,
        0,
        0,
        25,
        23,
        1,
        0,
        0,
        0,
        25,
        26,
        1,
        0,
        0,
        0,
        26,
        1,
        1,
        0,
        0,
        0,
        27,
        28,
        5,
        16,
        0,
        0,
        28,
        29,
        5,
        27,
        0,
        0,
        29,
        30,
        5,
        1,
        0,
        0,
        30,
        36,
        3,
        4,
        2,
        0,
        31,
        32,
        5,
        16,
        0,
        0,
        32,
        33,
        5,
        27,
        0,
        0,
        33,
        34,
        5,
        1,
        0,
        0,
        34,
        36,
        3,
        6,
        3,
        0,
        35,
        27,
        1,
        0,
        0,
        0,
        35,
        31,
        1,
        0,
        0,
        0,
        36,
        3,
        1,
        0,
        0,
        0,
        37,
        38,
        5,
        12,
        0,
        0,
        38,
        39,
        3,
        8,
        4,
        0,
        39,
        40,
        5,
        17,
        0,
        0,
        40,
        41,
        3,
        6,
        3,
        0,
        41,
        42,
        5,
        9,
        0,
        0,
        42,
        43,
        3,
        8,
        4,
        0,
        43,
        44,
        5,
        17,
        0,
        0,
        44,
        45,
        3,
        6,
        3,
        0,
        45,
        46,
        5,
        8,
        0,
        0,
        46,
        47,
        3,
        6,
        3,
        0,
        47,
        48,
        5,
        10,
        0,
        0,
        48,
        74,
        1,
        0,
        0,
        0,
        49,
        50,
        5,
        12,
        0,
        0,
        50,
        51,
        3,
        8,
        4,
        0,
        51,
        52,
        5,
        17,
        0,
        0,
        52,
        53,
        3,
        6,
        3,
        0,
        53,
        54,
        5,
        9,
        0,
        0,
        54,
        55,
        3,
        8,
        4,
        0,
        55,
        56,
        5,
        17,
        0,
        0,
        56,
        57,
        3,
        6,
        3,
        0,
        57,
        58,
        5,
        10,
        0,
        0,
        58,
        74,
        1,
        0,
        0,
        0,
        59,
        60,
        5,
        12,
        0,
        0,
        60,
        61,
        3,
        8,
        4,
        0,
        61,
        62,
        5,
        17,
        0,
        0,
        62,
        63,
        3,
        6,
        3,
        0,
        63,
        64,
        5,
        8,
        0,
        0,
        64,
        65,
        3,
        6,
        3,
        0,
        65,
        66,
        5,
        10,
        0,
        0,
        66,
        74,
        1,
        0,
        0,
        0,
        67,
        68,
        5,
        12,
        0,
        0,
        68,
        69,
        3,
        8,
        4,
        0,
        69,
        70,
        5,
        17,
        0,
        0,
        70,
        71,
        3,
        6,
        3,
        0,
        71,
        72,
        5,
        10,
        0,
        0,
        72,
        74,
        1,
        0,
        0,
        0,
        73,
        37,
        1,
        0,
        0,
        0,
        73,
        49,
        1,
        0,
        0,
        0,
        73,
        59,
        1,
        0,
        0,
        0,
        73,
        67,
        1,
        0,
        0,
        0,
        74,
        5,
        1,
        0,
        0,
        0,
        75,
        77,
        3,
        16,
        8,
        0,
        76,
        75,
        1,
        0,
        0,
        0,
        77,
        78,
        1,
        0,
        0,
        0,
        78,
        76,
        1,
        0,
        0,
        0,
        78,
        79,
        1,
        0,
        0,
        0,
        79,
        7,
        1,
        0,
        0,
        0,
        80,
        81,
        5,
        2,
        0,
        0,
        81,
        82,
        3,
        10,
        5,
        0,
        82,
        83,
        5,
        3,
        0,
        0,
        83,
        9,
        1,
        0,
        0,
        0,
        84,
        85,
        6,
        5,
        -1,
        0,
        85,
        92,
        3,
        12,
        6,
        0,
        86,
        92,
        3,
        14,
        7,
        0,
        87,
        88,
        5,
        2,
        0,
        0,
        88,
        89,
        3,
        10,
        5,
        0,
        89,
        90,
        5,
        3,
        0,
        0,
        90,
        92,
        1,
        0,
        0,
        0,
        91,
        84,
        1,
        0,
        0,
        0,
        91,
        86,
        1,
        0,
        0,
        0,
        91,
        87,
        1,
        0,
        0,
        0,
        92,
        98,
        1,
        0,
        0,
        0,
        93,
        94,
        10,
        4,
        0,
        0,
        94,
        95,
        5,
        5,
        0,
        0,
        95,
        97,
        3,
        10,
        5,
        5,
        96,
        93,
        1,
        0,
        0,
        0,
        97,
        100,
        1,
        0,
        0,
        0,
        98,
        96,
        1,
        0,
        0,
        0,
        98,
        99,
        1,
        0,
        0,
        0,
        99,
        11,
        1,
        0,
        0,
        0,
        100,
        98,
        1,
        0,
        0,
        0,
        101,
        102,
        5,
        27,
        0,
        0,
        102,
        103,
        5,
        6,
        0,
        0,
        103,
        109,
        3,
        18,
        9,
        0,
        104,
        105,
        3,
        18,
        9,
        0,
        105,
        106,
        5,
        6,
        0,
        0,
        106,
        107,
        5,
        27,
        0,
        0,
        107,
        109,
        1,
        0,
        0,
        0,
        108,
        101,
        1,
        0,
        0,
        0,
        108,
        104,
        1,
        0,
        0,
        0,
        109,
        13,
        1,
        0,
        0,
        0,
        110,
        111,
        5,
        25,
        0,
        0,
        111,
        112,
        5,
        6,
        0,
        0,
        112,
        118,
        3,
        20,
        10,
        0,
        113,
        114,
        3,
        20,
        10,
        0,
        114,
        115,
        5,
        6,
        0,
        0,
        115,
        116,
        5,
        25,
        0,
        0,
        116,
        118,
        1,
        0,
        0,
        0,
        117,
        110,
        1,
        0,
        0,
        0,
        117,
        113,
        1,
        0,
        0,
        0,
        118,
        15,
        1,
        0,
        0,
        0,
        119,
        120,
        5,
        7,
        0,
        0,
        120,
        121,
        5,
        26,
        0,
        0,
        121,
        137,
        5,
        4,
        0,
        0,
        122,
        123,
        5,
        15,
        0,
        0,
        123,
        124,
        5,
        26,
        0,
        0,
        124,
        137,
        5,
        4,
        0,
        0,
        125,
        126,
        5,
        13,
        0,
        0,
        126,
        127,
        5,
        25,
        0,
        0,
        127,
        137,
        5,
        4,
        0,
        0,
        128,
        129,
        5,
        14,
        0,
        0,
        129,
        130,
        5,
        25,
        0,
        0,
        130,
        137,
        5,
        4,
        0,
        0,
        131,
        132,
        5,
        11,
        0,
        0,
        132,
        133,
        5,
        27,
        0,
        0,
        133,
        137,
        5,
        4,
        0,
        0,
        134,
        135,
        5,
        18,
        0,
        0,
        135,
        137,
        5,
        4,
        0,
        0,
        136,
        119,
        1,
        0,
        0,
        0,
        136,
        122,
        1,
        0,
        0,
        0,
        136,
        125,
        1,
        0,
        0,
        0,
        136,
        128,
        1,
        0,
        0,
        0,
        136,
        131,
        1,
        0,
        0,
        0,
        136,
        134,
        1,
        0,
        0,
        0,
        137,
        17,
        1,
        0,
        0,
        0,
        138,
        142,
        3,
        20,
        10,
        0,
        139,
        142,
        5,
        22,
        0,
        0,
        140,
        142,
        5,
        23,
        0,
        0,
        141,
        138,
        1,
        0,
        0,
        0,
        141,
        139,
        1,
        0,
        0,
        0,
        141,
        140,
        1,
        0,
        0,
        0,
        142,
        19,
        1,
        0,
        0,
        0,
        143,
        147,
        5,
        19,
        0,
        0,
        144,
        147,
        5,
        20,
        0,
        0,
        145,
        147,
        5,
        21,
        0,
        0,
        146,
        143,
        1,
        0,
        0,
        0,
        146,
        144,
        1,
        0,
        0,
        0,
        146,
        145,
        1,
        0,
        0,
        0,
        147,
        21,
        1,
        0,
        0,
        0,
        11,
        25,
        35,
        73,
        78,
        91,
        98,
        108,
        117,
        136,
        141,
        146,
    ]


class ILATsmParser(Parser):
    grammarFileName = "ILATsm.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = ["<INVALID>", "':'", "'('", "')'", "';'"]

    symbolicNames = [
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "LOGIC_OP",
        "COMPARE_OP",
        "CLEAR_FLAG",
        "ELSE",
        "ELSEIF",
        "ENDIF",
        "GOTO",
        "IF",
        "INCREMENT_COUNTER",
        "RESET_COUNTER",
        "SET_FLAG",
        "STATE",
        "THEN",
        "TRIGGER",
        "INTEGER_VAL",
        "HEX_VAL",
        "BINARY_VAL",
        "ENUM_VAL",
        "ENUM_STRING_VAL",
        "LINE_COMMENT",
        "COUNTER_NAME",
        "FLAG_NAME",
        "IDENTIFIER",
        "PROBE_BUS_RANGE",
        "TRUE",
        "FALSE",
        "WS",
    ]

    RULE_ila_tsm = 0
    RULE_state_block = 1
    RULE_if_condition = 2
    RULE_actions = 3
    RULE_full_condition = 4
    RULE_condition = 5
    RULE_probe_match = 6
    RULE_counter_match = 7
    RULE_action = 8
    RULE_probe_number = 9
    RULE_number = 10

    ruleNames = [
        "ila_tsm",
        "state_block",
        "if_condition",
        "actions",
        "full_condition",
        "condition",
        "probe_match",
        "counter_match",
        "action",
        "probe_number",
        "number",
    ]

    EOF = Token.EOF
    T__0 = 1
    T__1 = 2
    T__2 = 3
    T__3 = 4
    LOGIC_OP = 5
    COMPARE_OP = 6
    CLEAR_FLAG = 7
    ELSE = 8
    ELSEIF = 9
    ENDIF = 10
    GOTO = 11
    IF = 12
    INCREMENT_COUNTER = 13
    RESET_COUNTER = 14
    SET_FLAG = 15
    STATE = 16
    THEN = 17
    TRIGGER = 18
    INTEGER_VAL = 19
    HEX_VAL = 20
    BINARY_VAL = 21
    ENUM_VAL = 22
    ENUM_STRING_VAL = 23
    LINE_COMMENT = 24
    COUNTER_NAME = 25
    FLAG_NAME = 26
    IDENTIFIER = 27
    PROBE_BUS_RANGE = 28
    TRUE = 29
    FALSE = 30
    WS = 31

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class Ila_tsmContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def state_block(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ILATsmParser.State_blockContext)
            else:
                return self.getTypedRuleContext(ILATsmParser.State_blockContext, i)

        def getRuleIndex(self):
            return ILATsmParser.RULE_ila_tsm

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIla_tsm"):
                return visitor.visitIla_tsm(self)
            else:
                return visitor.visitChildren(self)

    def ila_tsm(self):
        localctx = ILATsmParser.Ila_tsmContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_ila_tsm)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 23
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 22
                self.state_block()
                self.state = 25
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la == 16):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class State_blockContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ILATsmParser.RULE_state_block

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class State_no_ifContext(State_blockContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.State_blockContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STATE(self):
            return self.getToken(ILATsmParser.STATE, 0)

        def IDENTIFIER(self):
            return self.getToken(ILATsmParser.IDENTIFIER, 0)

        def actions(self):
            return self.getTypedRuleContext(ILATsmParser.ActionsContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitState_no_if"):
                return visitor.visitState_no_if(self)
            else:
                return visitor.visitChildren(self)

    class State_ifContext(State_blockContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.State_blockContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def STATE(self):
            return self.getToken(ILATsmParser.STATE, 0)

        def IDENTIFIER(self):
            return self.getToken(ILATsmParser.IDENTIFIER, 0)

        def if_condition(self):
            return self.getTypedRuleContext(ILATsmParser.If_conditionContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitState_if"):
                return visitor.visitState_if(self)
            else:
                return visitor.visitChildren(self)

    def state_block(self):
        localctx = ILATsmParser.State_blockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_state_block)
        try:
            self.state = 35
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 1, self._ctx)
            if la_ == 1:
                localctx = ILATsmParser.State_ifContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 27
                self.match(ILATsmParser.STATE)
                self.state = 28
                self.match(ILATsmParser.IDENTIFIER)
                self.state = 29
                self.match(ILATsmParser.T__0)
                self.state = 30
                self.if_condition()
                pass

            elif la_ == 2:
                localctx = ILATsmParser.State_no_ifContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 31
                self.match(ILATsmParser.STATE)
                self.state = 32
                self.match(ILATsmParser.IDENTIFIER)
                self.state = 33
                self.match(ILATsmParser.T__0)
                self.state = 34
                self.actions()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class If_conditionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IF(self):
            return self.getToken(ILATsmParser.IF, 0)

        def full_condition(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ILATsmParser.Full_conditionContext)
            else:
                return self.getTypedRuleContext(ILATsmParser.Full_conditionContext, i)

        def THEN(self, i: int = None):
            if i is None:
                return self.getTokens(ILATsmParser.THEN)
            else:
                return self.getToken(ILATsmParser.THEN, i)

        def actions(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ILATsmParser.ActionsContext)
            else:
                return self.getTypedRuleContext(ILATsmParser.ActionsContext, i)

        def ELSEIF(self):
            return self.getToken(ILATsmParser.ELSEIF, 0)

        def ELSE(self):
            return self.getToken(ILATsmParser.ELSE, 0)

        def ENDIF(self):
            return self.getToken(ILATsmParser.ENDIF, 0)

        def getRuleIndex(self):
            return ILATsmParser.RULE_if_condition

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitIf_condition"):
                return visitor.visitIf_condition(self)
            else:
                return visitor.visitChildren(self)

    def if_condition(self):
        localctx = ILATsmParser.If_conditionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_if_condition)
        try:
            self.state = 73
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 37
                self.match(ILATsmParser.IF)
                self.state = 38
                self.full_condition()
                self.state = 39
                self.match(ILATsmParser.THEN)
                self.state = 40
                self.actions()
                self.state = 41
                self.match(ILATsmParser.ELSEIF)
                self.state = 42
                self.full_condition()
                self.state = 43
                self.match(ILATsmParser.THEN)
                self.state = 44
                self.actions()
                self.state = 45
                self.match(ILATsmParser.ELSE)
                self.state = 46
                self.actions()
                self.state = 47
                self.match(ILATsmParser.ENDIF)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 49
                self.match(ILATsmParser.IF)
                self.state = 50
                self.full_condition()
                self.state = 51
                self.match(ILATsmParser.THEN)
                self.state = 52
                self.actions()
                self.state = 53
                self.match(ILATsmParser.ELSEIF)
                self.state = 54
                self.full_condition()
                self.state = 55
                self.match(ILATsmParser.THEN)
                self.state = 56
                self.actions()
                self.state = 57
                self.match(ILATsmParser.ENDIF)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 59
                self.match(ILATsmParser.IF)
                self.state = 60
                self.full_condition()
                self.state = 61
                self.match(ILATsmParser.THEN)
                self.state = 62
                self.actions()
                self.state = 63
                self.match(ILATsmParser.ELSE)
                self.state = 64
                self.actions()
                self.state = 65
                self.match(ILATsmParser.ENDIF)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 67
                self.match(ILATsmParser.IF)
                self.state = 68
                self.full_condition()
                self.state = 69
                self.match(ILATsmParser.THEN)
                self.state = 70
                self.actions()
                self.state = 71
                self.match(ILATsmParser.ENDIF)
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ActionsContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def action(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ILATsmParser.ActionContext)
            else:
                return self.getTypedRuleContext(ILATsmParser.ActionContext, i)

        def getRuleIndex(self):
            return ILATsmParser.RULE_actions

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitActions"):
                return visitor.visitActions(self)
            else:
                return visitor.visitChildren(self)

    def actions(self):
        localctx = ILATsmParser.ActionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_actions)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 75
                self.action()
                self.state = 78
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3F) == 0 and ((1 << _la) & 321664) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Full_conditionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def condition(self):
            return self.getTypedRuleContext(ILATsmParser.ConditionContext, 0)

        def getRuleIndex(self):
            return ILATsmParser.RULE_full_condition

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitFull_condition"):
                return visitor.visitFull_condition(self)
            else:
                return visitor.visitChildren(self)

    def full_condition(self):
        localctx = ILATsmParser.Full_conditionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_full_condition)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 80
            self.match(ILATsmParser.T__1)
            self.state = 81
            self.condition(0)
            self.state = 82
            self.match(ILATsmParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConditionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ILATsmParser.RULE_condition

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Condition_counter_matchContext(ConditionContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def counter_match(self):
            return self.getTypedRuleContext(ILATsmParser.Counter_matchContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCondition_counter_match"):
                return visitor.visitCondition_counter_match(self)
            else:
                return visitor.visitChildren(self)

    class Condition_parenContext(ConditionContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def condition(self):
            return self.getTypedRuleContext(ILATsmParser.ConditionContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCondition_paren"):
                return visitor.visitCondition_paren(self)
            else:
                return visitor.visitChildren(self)

    class Condition_logic_opContext(ConditionContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def condition(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(ILATsmParser.ConditionContext)
            else:
                return self.getTypedRuleContext(ILATsmParser.ConditionContext, i)

        def LOGIC_OP(self):
            return self.getToken(ILATsmParser.LOGIC_OP, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCondition_logic_op"):
                return visitor.visitCondition_logic_op(self)
            else:
                return visitor.visitChildren(self)

    class Condition_probe_matchContext(ConditionContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.ConditionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def probe_match(self):
            return self.getTypedRuleContext(ILATsmParser.Probe_matchContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCondition_probe_match"):
                return visitor.visitCondition_probe_match(self)
            else:
                return visitor.visitChildren(self)

    def condition(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = ILATsmParser.ConditionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 10
        self.enterRecursionRule(localctx, 10, self.RULE_condition, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 91
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 4, self._ctx)
            if la_ == 1:
                localctx = ILATsmParser.Condition_probe_matchContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 85
                self.probe_match()
                pass

            elif la_ == 2:
                localctx = ILATsmParser.Condition_counter_matchContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 86
                self.counter_match()
                pass

            elif la_ == 3:
                localctx = ILATsmParser.Condition_parenContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 87
                self.match(ILATsmParser.T__1)
                self.state = 88
                self.condition(0)
                self.state = 89
                self.match(ILATsmParser.T__2)
                pass

            self._ctx.stop = self._input.LT(-1)
            self.state = 98
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 5, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = ILATsmParser.Condition_logic_opContext(
                        self, ILATsmParser.ConditionContext(self, _parentctx, _parentState)
                    )
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_condition)
                    self.state = 93
                    if not self.precpred(self._ctx, 4):
                        from antlr4.error.Errors import FailedPredicateException

                        raise FailedPredicateException(self, "self.precpred(self._ctx, 4)")
                    self.state = 94
                    self.match(ILATsmParser.LOGIC_OP)
                    self.state = 95
                    self.condition(5)
                self.state = 100
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 5, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class Probe_matchContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ILATsmParser.IDENTIFIER, 0)

        def COMPARE_OP(self):
            return self.getToken(ILATsmParser.COMPARE_OP, 0)

        def probe_number(self):
            return self.getTypedRuleContext(ILATsmParser.Probe_numberContext, 0)

        def getRuleIndex(self):
            return ILATsmParser.RULE_probe_match

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitProbe_match"):
                return visitor.visitProbe_match(self)
            else:
                return visitor.visitChildren(self)

    def probe_match(self):
        localctx = ILATsmParser.Probe_matchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_probe_match)
        try:
            self.state = 108
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [27]:
                self.enterOuterAlt(localctx, 1)
                self.state = 101
                self.match(ILATsmParser.IDENTIFIER)
                self.state = 102
                self.match(ILATsmParser.COMPARE_OP)
                self.state = 103
                self.probe_number()
                pass
            elif token in [19, 20, 21, 22, 23]:
                self.enterOuterAlt(localctx, 2)
                self.state = 104
                self.probe_number()
                self.state = 105
                self.match(ILATsmParser.COMPARE_OP)
                self.state = 106
                self.match(ILATsmParser.IDENTIFIER)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Counter_matchContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def COUNTER_NAME(self):
            return self.getToken(ILATsmParser.COUNTER_NAME, 0)

        def COMPARE_OP(self):
            return self.getToken(ILATsmParser.COMPARE_OP, 0)

        def number(self):
            return self.getTypedRuleContext(ILATsmParser.NumberContext, 0)

        def getRuleIndex(self):
            return ILATsmParser.RULE_counter_match

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitCounter_match"):
                return visitor.visitCounter_match(self)
            else:
                return visitor.visitChildren(self)

    def counter_match(self):
        localctx = ILATsmParser.Counter_matchContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_counter_match)
        try:
            self.state = 117
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [25]:
                self.enterOuterAlt(localctx, 1)
                self.state = 110
                self.match(ILATsmParser.COUNTER_NAME)
                self.state = 111
                self.match(ILATsmParser.COMPARE_OP)
                self.state = 112
                self.number()
                pass
            elif token in [19, 20, 21]:
                self.enterOuterAlt(localctx, 2)
                self.state = 113
                self.number()
                self.state = 114
                self.match(ILATsmParser.COMPARE_OP)
                self.state = 115
                self.match(ILATsmParser.COUNTER_NAME)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ActionContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ILATsmParser.RULE_action

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Action_set_flagContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def SET_FLAG(self):
            return self.getToken(ILATsmParser.SET_FLAG, 0)

        def FLAG_NAME(self):
            return self.getToken(ILATsmParser.FLAG_NAME, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_set_flag"):
                return visitor.visitAction_set_flag(self)
            else:
                return visitor.visitChildren(self)

    class Action_reset_counterContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def RESET_COUNTER(self):
            return self.getToken(ILATsmParser.RESET_COUNTER, 0)

        def COUNTER_NAME(self):
            return self.getToken(ILATsmParser.COUNTER_NAME, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_reset_counter"):
                return visitor.visitAction_reset_counter(self)
            else:
                return visitor.visitChildren(self)

    class Action_gotoContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def GOTO(self):
            return self.getToken(ILATsmParser.GOTO, 0)

        def IDENTIFIER(self):
            return self.getToken(ILATsmParser.IDENTIFIER, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_goto"):
                return visitor.visitAction_goto(self)
            else:
                return visitor.visitChildren(self)

    class Action_triggerContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def TRIGGER(self):
            return self.getToken(ILATsmParser.TRIGGER, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_trigger"):
                return visitor.visitAction_trigger(self)
            else:
                return visitor.visitChildren(self)

    class Action_inc_counterContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INCREMENT_COUNTER(self):
            return self.getToken(ILATsmParser.INCREMENT_COUNTER, 0)

        def COUNTER_NAME(self):
            return self.getToken(ILATsmParser.COUNTER_NAME, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_inc_counter"):
                return visitor.visitAction_inc_counter(self)
            else:
                return visitor.visitChildren(self)

    class Action_clear_flagContext(ActionContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.ActionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def CLEAR_FLAG(self):
            return self.getToken(ILATsmParser.CLEAR_FLAG, 0)

        def FLAG_NAME(self):
            return self.getToken(ILATsmParser.FLAG_NAME, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitAction_clear_flag"):
                return visitor.visitAction_clear_flag(self)
            else:
                return visitor.visitChildren(self)

    def action(self):
        localctx = ILATsmParser.ActionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_action)
        try:
            self.state = 136
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [7]:
                localctx = ILATsmParser.Action_clear_flagContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 119
                self.match(ILATsmParser.CLEAR_FLAG)
                self.state = 120
                self.match(ILATsmParser.FLAG_NAME)
                self.state = 121
                self.match(ILATsmParser.T__3)
                pass
            elif token in [15]:
                localctx = ILATsmParser.Action_set_flagContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 122
                self.match(ILATsmParser.SET_FLAG)
                self.state = 123
                self.match(ILATsmParser.FLAG_NAME)
                self.state = 124
                self.match(ILATsmParser.T__3)
                pass
            elif token in [13]:
                localctx = ILATsmParser.Action_inc_counterContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 125
                self.match(ILATsmParser.INCREMENT_COUNTER)
                self.state = 126
                self.match(ILATsmParser.COUNTER_NAME)
                self.state = 127
                self.match(ILATsmParser.T__3)
                pass
            elif token in [14]:
                localctx = ILATsmParser.Action_reset_counterContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 128
                self.match(ILATsmParser.RESET_COUNTER)
                self.state = 129
                self.match(ILATsmParser.COUNTER_NAME)
                self.state = 130
                self.match(ILATsmParser.T__3)
                pass
            elif token in [11]:
                localctx = ILATsmParser.Action_gotoContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 131
                self.match(ILATsmParser.GOTO)
                self.state = 132
                self.match(ILATsmParser.IDENTIFIER)
                self.state = 133
                self.match(ILATsmParser.T__3)
                pass
            elif token in [18]:
                localctx = ILATsmParser.Action_triggerContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 134
                self.match(ILATsmParser.TRIGGER)
                self.state = 135
                self.match(ILATsmParser.T__3)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Probe_numberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ILATsmParser.RULE_probe_number

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Enum_string_valContext(Probe_numberContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.Probe_numberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ENUM_STRING_VAL(self):
            return self.getToken(ILATsmParser.ENUM_STRING_VAL, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitEnum_string_val"):
                return visitor.visitEnum_string_val(self)
            else:
                return visitor.visitChildren(self)

    class Probe_valContext(Probe_numberContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.Probe_numberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def number(self):
            return self.getTypedRuleContext(ILATsmParser.NumberContext, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitProbe_val"):
                return visitor.visitProbe_val(self)
            else:
                return visitor.visitChildren(self)

    class Enum_valContext(Probe_numberContext):
        def __init__(
            self, parser, ctx: ParserRuleContext
        ):  # actually a ILATsmParser.Probe_numberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def ENUM_VAL(self):
            return self.getToken(ILATsmParser.ENUM_VAL, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitEnum_val"):
                return visitor.visitEnum_val(self)
            else:
                return visitor.visitChildren(self)

    def probe_number(self):
        localctx = ILATsmParser.Probe_numberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_probe_number)
        try:
            self.state = 141
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [19, 20, 21]:
                localctx = ILATsmParser.Probe_valContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 138
                self.number()
                pass
            elif token in [22]:
                localctx = ILATsmParser.Enum_valContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 139
                self.match(ILATsmParser.ENUM_VAL)
                pass
            elif token in [23]:
                localctx = ILATsmParser.Enum_string_valContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 140
                self.match(ILATsmParser.ENUM_STRING_VAL)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def getRuleIndex(self):
            return ILATsmParser.RULE_number

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)

    class Int_valContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.NumberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INTEGER_VAL(self):
            return self.getToken(ILATsmParser.INTEGER_VAL, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitInt_val"):
                return visitor.visitInt_val(self)
            else:
                return visitor.visitChildren(self)

    class Bin_valContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.NumberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def BINARY_VAL(self):
            return self.getToken(ILATsmParser.BINARY_VAL, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitBin_val"):
                return visitor.visitBin_val(self)
            else:
                return visitor.visitChildren(self)

    class Hex_valContext(NumberContext):
        def __init__(self, parser, ctx: ParserRuleContext):  # actually a ILATsmParser.NumberContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def HEX_VAL(self):
            return self.getToken(ILATsmParser.HEX_VAL, 0)

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitHex_val"):
                return visitor.visitHex_val(self)
            else:
                return visitor.visitChildren(self)

    def number(self):
        localctx = ILATsmParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_number)
        try:
            self.state = 146
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [19]:
                localctx = ILATsmParser.Int_valContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 143
                self.match(ILATsmParser.INTEGER_VAL)
                pass
            elif token in [20]:
                localctx = ILATsmParser.Hex_valContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 144
                self.match(ILATsmParser.HEX_VAL)
                pass
            elif token in [21]:
                localctx = ILATsmParser.Bin_valContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 145
                self.match(ILATsmParser.BINARY_VAL)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[5] = self.condition_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def condition_sempred(self, localctx: ConditionContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 4)



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from enum import Enum, EnumMeta
from io import TextIOBase
from collections import defaultdict
from dataclasses import dataclass, asdict
from os import path
from pathlib import Path
from pprint import pformat
from re import compile as re_compile
from typing import Union, Dict, List, Optional, Tuple, NewType, Set, Any

from chipscopy.api import CoreType
from chipscopy.utils import Enum2StrEncoder


class LtxPath:
    TOP_LEVEL_STR = ""
    # Path string value for top-level static partition is "".

    def __init__(self, path_str: str, source: str):
        """

        Args:
            path_str: Partition cell name or "" for top-level.
            source: Ltx source. Filename or other name, if filename is not known.
        """
        self._str: str = path_str
        self._segments: List[str] = path_str.split("/")
        self._source: str = source

    def path_segments(self) -> List[str]:
        return self._segments

    def get_source_name(self):
        return self._source

    def is_top_level(self) -> bool:
        return self._str == ""

    def is_inside_of(self, other) -> bool:
        # True if "self" is inside "other" or identical to "other".
        # E.g. Partition path "a/b/c" is inside partition path "a/b".
        # E.g. Partition path "a/b/e" is inside partition path "a/b/e".
        # Top level is special case. Re-programmable partitions are not considered
        #  to be part of top-level static partition.
        if self.is_top_level():
            return other.is_top_level()
        if len(other.path_segments()) > len(self._segments):
            return False
        for seg, other_seg in zip(self._segments, other.path_segments()):
            if seg != other_seg:
                return False
        return True

    def __repr__(self) -> str:
        return pformat((self._str, self._source))

    def __str__(self) -> str:
        return self._str

    def __eq__(self, other):
        return self._str == str(other)

    def __hash__(self):
        return hash(self._str)


@dataclass(frozen=True)
class LtxProbe:
    """Represents a single probe in the LTX file"""

    name: str
    direction: str
    probe_type: str
    port_index: int
    port_msb_index: int
    port_lsb_index: int
    is_bus: bool
    bus_left_index: int
    bus_right_index: int
    enum_def: Optional[EnumMeta]


@dataclass(frozen=True)
class LtxCore:
    core_type: CoreType
    cell_name: str
    uuid: str
    debug_hub_address: int
    probes: [LtxProbe]
    upstream_cell_names: List[str]
    """ E.g. A VIO debug core would list the connected debug_hub, as an upstream core."""
    partition: LtxPath

    def is_inside_static_partition(self) -> bool:
        """True of the core is inside static partition."""
        return self.partition.is_top_level()

    def as_json_dict(self) -> Dict[str, Any]:
        j_dict = asdict(self)
        j_dict["partition"] = str(self.partition)
        return j_dict

    def __str__(self) -> str:
        return pformat(self.__dict__)


@dataclass(frozen=True)
class LtxStreamRef:
    """
    Reference to a debug core stream interface.
    A debug core has one or more control/data streams which connect to other debug cores.
    E.g. A debug hub may have multiple downstream connections,
    one for each connected VIO and ILA debug core.
    """

    cell_name: str
    uuid: int
    stream_index: int


ParentEnums = NewType("ParentEnums", Dict[int, Dict[str, str]])


class Ltx:
    __LTX_IP_NAME_TO_CORE_TYPE = {
        "axi_dbg_hub": CoreType.AXI_DEBUG_HUB,
        "axis_ila": CoreType.AXIS_ILA,
        "axis_vio": CoreType.AXIS_VIO,
        "trace": CoreType.AXIS_TRACE,
        "hw_trace": CoreType.AXIS_TRACE,
    }

    __NULL_UUID = "0" * 32
    __net_bus_index_re = None
    __subnet_bus_index_re = None

    def __init__(self, source_name=""):
        self._source_name: str = source_name
        self._cores: {CoreType, [LtxCore]} = defaultdict(list)
        self._debug_hub_addresses: [int] = []
        self._downstream_refs: {str: List[LtxStreamRef]} = defaultdict(list)
        """
            key: upstream_cell_name. value: list of downstream LtxStreamRef
            E.g. an entry for a debug_hub core, would list other debug cores connected to it.
        """
        self._partitions: Set[LtxPath] = set()
        """Empty partitions have no cores, but still important to keep track of."""
        self._empty_partitions: set[LtxPath] = set()

    def is_top_level(self) -> bool:
        return LtxPath(LtxPath.TOP_LEVEL_STR, "") in self._partitions

    def get_source_names(self):
        # Names of source files. For DFX designs may be more than one.
        names = [partition.get_source_name() for partition in self._partitions]
        return names

    def is_inside_ltx_partitions(self, partition: LtxPath) -> bool:
        inside = any(partition.is_inside_of(part) for part in self._partitions)
        return inside

    def replace_partition_cores(self, other):
        other_ltx: Ltx = other
        # Reset source name, since it is now a combined Ltx object with data from multiple files.
        self._source_name = ""
        # Remove cores which exist in "other_ltx" partitions.
        for core_type, core_list in self._cores.items():
            self._cores[core_type] = [
                core for core in core_list if not other_ltx.is_inside_ltx_partitions(core.partition)
            ]
        # Remove empty partitions which has partitions in "other_ltx".
        self._empty_partitions = {
            ep for ep in self._empty_partitions if not other_ltx.is_inside_ltx_partitions(ep)
        }
        # Add cores from "other_ltx"
        for core_type, core_list in other_ltx._cores.items():
            for core in core_list:
                self._cores[core_type].append(core)
        # Add empty partitions from "other_lts"
        self._empty_partitions.update(other_ltx._empty_partitions)

    def get_debug_hub_addresses(self) -> [int]:
        return self._debug_hub_addresses

    def get_core(self, core_type: CoreType, uuid: str = "") -> Optional[LtxCore]:
        """Return first matching core, or None."""
        cores = [core for core in self._cores[core_type] if core.uuid.upper() == uuid.upper()]
        return cores[0] if cores else None

    def get_downstream_refs(self, upstream_core_name: str) -> List[LtxStreamRef]:
        """
        Returns list of LtxStreamRef, which lists which debug cores the upstream core is connected to.
        The list is not ordered.
        E.g. given a debug_hub cell_name, the function returns a list of connected debug cores.
        """
        return self._downstream_refs.get(upstream_core_name, [])

    def get_cores(self, core_type: CoreType) -> [LtxCore]:
        return self._cores[core_type]

    def parse_file(self, probes_file: Union[Path, str]) -> None:
        with open(probes_file, "r") as json_file:
            self.parse_ltx(json_file)

    def parse_ltx(self, fh: TextIOBase) -> None:
        """
        A Ltx file with no "cellName" items, is a top-level Ltx file.
        """
        cell_name_item_count = 0
        json_data = json.load(fh)

        ltx_root = json_data.get("ltx_root", None)
        ltx_data = ltx_root.get("ltx_data", []) if ltx_root else []

        for ltx_data_item in ltx_data:
            cell_name = ltx_data_item.get("cellName", None)
            path = None
            if cell_name:
                # "CellName" item only occurs for RMs in partial Ltx files.
                # The RM may have zero or more debug cores.
                cell_name_item_count += 1
                path = LtxPath(cell_name, self._source_name)
                self._partitions.add(path)

            parent_id_to_enum = {}
            if "parent_nets" in ltx_data_item:
                parent_id_to_enum = self._read_parent_net_enums(ltx_data_item["parent_nets"])

            debug_cores = ltx_data_item.get("debug_cores", [])
            for debug_core in debug_cores:
                core = Ltx._read_core(debug_core, parent_id_to_enum, self._source_name)
                if not core:
                    continue
                self._cores[core.core_type].append(core)

            if not debug_cores and path:
                self._empty_partitions.add(path)

        if cell_name_item_count == 0:
            # RM Ltx files have 'cellName' items.
            # Full Ltx files, do not have 'cellName' items.
            self._partitions.add(LtxPath(LtxPath.TOP_LEVEL_STR, self._source_name))

        self.post_process()

    def post_process(self):
        """
        Update Ltx object data members:
            - debug_hub addresses
            - core downstream refs
            - partitions
        """
        self._debug_hub_addresses = []
        self._downstream_refs = defaultdict(list)
        self._partitions = {partition for partition in self._partitions if partition.is_top_level()}
        self._partitions.update(self._empty_partitions)

        for core_list in self._cores.values():
            for core in core_list:
                # Update core partition path.
                self._partitions.add(core.partition)

                # Update core down-stream refs.
                for idx, up_stream_cell_names in enumerate(core.upstream_cell_names):
                    self._downstream_refs[up_stream_cell_names].append(
                        LtxStreamRef(core.cell_name, core.uuid, idx)
                    )

        # Sort cores, to preserve order for tests, etc.
        for core_list in self._cores.values():
            core_list.sort(key=lambda x: x.cell_name)

        # Update debug hub address(es).
        self._debug_hub_addresses.extend(
            int(core.debug_hub_address, 16) for core in self._cores[CoreType.AXI_DEBUG_HUB]
        )

    def verify_port_width_by_type(
        self, core_type: CoreType, uuid: str, width_by_type: Dict[str, List[int]]
    ) -> None:
        mismatch_msg = "Mismatch between LTX file(s) and device debug core.\n"
        core = self.get_core(core_type, uuid)
        if not core:
            raise KeyError(
                mismatch_msg
                + f"Device {core_type.value} core with UUID {uuid} is not found in the LTX file."
            )

        for probe in core.probes:
            port_widths = width_by_type.get(probe.probe_type, [])
            port_width = port_widths[probe.port_index] if len(port_widths) > probe.port_index else 0
            if port_width == 0:
                raise KeyError(
                    mismatch_msg
                    + f'LTX core "{core.cell_name}" probe "{probe.name}" of type {probe.probe_type} '
                    f'refers to core port index "{probe.port_index}", '
                    f"which does not exist in the debug core for probe type {probe.probe_type}."
                )
            if port_width <= probe.port_msb_index:
                raise ValueError(
                    mismatch_msg
                    + f'LTX core "{core.cell_name}" probe "{probe.name}" of type {probe.probe_type} '
                    f"has invalid msb index:{probe.port_msb_index}.\n The port width is {port_width}"
                )

    @staticmethod
    def _read_parent_net_enums(parent_nets: []) -> ParentEnums:
        """State values are returned as binary strings, so the bit_width is obvious.
        Return type: {parent_id: {state_name, state_binary_str_value}}
        """
        pid_to_enums = {}
        for pnet in parent_nets:
            pid_states = {}
            for sr in pnet.get("state_reg", []):
                pid_states.update((state["name"], state["val"]) for state in sr.get("states", []))
            if pid_states:
                pid_to_enums[pnet["id"]] = pid_states
        return pid_to_enums

    @staticmethod
    def _get_enum_def(
        probe_name: str, parent_ids: [int], parent_enums: ParentEnums, probe_width: int
    ) -> Optional[Enum]:
        if not parent_ids:
            return None
        p_enum = parent_enums.get(parent_ids[0], None)
        if not all([len(val) == probe_width for val in p_enum.values()]):
            # Old LTX files have bad enums, for non-important probes. Skip enum.
            return None
        enum_vals = {state: int(val, 2) for state, val in p_enum.items()}
        return Enum(probe_name, enum_vals)

    @staticmethod
    def _read_core(
        in_core: {}, parent_enums: ParentEnums, source_name: str
    ) -> Union[LtxCore, None]:
        partition_str = in_core.get("reconfigTop", "")
        partition = LtxPath(partition_str, source_name)
        pins = in_core.get("pins", [])
        probes = [Ltx._read_pin(pin, parent_enums) for pin in pins]
        probes = [probe for probe in probes if probe]
        ip_name = in_core.get("ipName", "")
        if ip_name == "axi_dbg_hub" and "address_info" in in_core:
            addr = in_core["address_info"].get("offset", "")
        else:
            addr = ""

        core_type = Ltx.__LTX_IP_NAME_TO_CORE_TYPE.get(ip_name, "")
        cell_name = in_core.get("name", "")
        uuid = in_core.get("uuid", Ltx.__NULL_UUID).upper()
        upstream = in_core.get("master", "")
        upstream_cell_names = upstream.split(":") if upstream else []

        # todo syntax/semantic checks?
        if not core_type:
            return None
        return LtxCore(core_type, cell_name, uuid, addr, probes, upstream_cell_names, partition)

    @staticmethod
    def read_nets(nets: []) -> Tuple[bool, int, int, List[int]]:
        """
        Returns (name, left index, right index, [parent_ids])
        """
        if not nets:
            return False, 0, 0, []

        parent_ids = [net.get("parentNetId", None) for net in nets]
        parent_ids = [p_id for p_id in parent_ids if p_id is not None]

        left_name = nets[0].get("name", "")
        right_name = nets[-1].get("name", "")

        if not Ltx.__net_bus_index_re:
            Ltx.__net_bus_index_re = re_compile(r"(.+)\[(-?\d+)]$")
        left_m = Ltx.__net_bus_index_re.match(left_name)
        right_m = Ltx.__net_bus_index_re.match(right_name)
        if not left_m or not right_m:
            return False, 0, 0, parent_ids
        return left_m.group(1), int(left_m.group(2)), int(right_m.group(2)), parent_ids

    @staticmethod
    def read_subnets(subnets: []) -> Tuple[int, int, List[int]]:
        """
        If a bus returns (left index, right index, [parent_ids])
        """
        if not subnets:
            return 0, 0, []

        parent_ids = [subnet.get("parentNetId", None) for subnet in subnets]
        parent_ids = [p_id for p_id in parent_ids if p_id is not None]

        left_name = subnets[0].get("name", "")
        right_name = subnets[-1].get("name", "")

        if not Ltx.__subnet_bus_index_re:
            Ltx.__subnet_bus_index_re = re_compile(r".+\[(-?\d+)]$")
        left_m = Ltx.__subnet_bus_index_re.match(left_name)
        right_m = Ltx.__subnet_bus_index_re.match(right_name)
        if not left_m or not right_m:
            return 0, 0, parent_ids
        return int(left_m.group(1)), int(right_m.group(1)), parent_ids

    @staticmethod
    def _read_pin(pin: {}, parent_enums: ParentEnums) -> LtxProbe:
        nets = pin.get("nets", [])
        if len(nets) == 1:
            name = nets[0].get("name", "") if nets else ""
            subnets = nets[0].get("subnets", [])
            bus_left_index, bus_right_index, parent_ids = Ltx.read_subnets(subnets)
        else:
            # Some state machine registers are split up in multiple nets.
            name, bus_left_index, bus_right_index, parent_ids = Ltx.read_nets(nets)

        probe_type = pin.get("type", "")
        is_bus = pin.get("isVector", False)
        port_index = pin.get("portIndex", 0)
        # The LTX file has port right/left switched.
        port_msb_index = pin.get("rightIndex", 0)
        port_lsb_index = pin.get("leftIndex", 0)
        direction = pin.get("direction", "")
        probe_def = Ltx._get_enum_def(
            name, parent_ids, parent_enums, abs(bus_left_index - bus_right_index) + 1
        )
        return LtxProbe(
            name,
            direction,
            probe_type,
            port_index,
            port_msb_index,
            port_lsb_index,
            is_bus,
            bus_left_index,
            bus_right_index,
            probe_def,
        )

    def __str__(self) -> str:
        dump = "\ndebug_hub_addresses: " + ", ".join(
            [hex(addr) for addr in self._debug_hub_addresses]
        )
        for core_type, cores in self._cores.items():
            underline = "=" * (len(core_type.__str__()) + 4)
            for idx, core in enumerate(cores):
                dump += f"\n\n{core_type}:{idx:3}\n{underline}\n"
                dump += core.__str__() + "\n"

        return dump

    def __repr__(self) -> str:
        cores = defaultdict(list)
        for core_type, core_list in self._cores.items():
            for core in core_list:
                cores[str(core_type)].append(core.as_json_dict())
        downstream_refs = defaultdict(list)
        for up, down_list in self._downstream_refs.items():
            for down in down_list:
                downstream_refs[up].append(asdict(down))

        partitions = [str(top) for top in self._partitions]
        partitions.sort()
        ret_dict = {
            "debug_hub_addresses": [f"0x{addr:x}" for addr in self._debug_hub_addresses],
            "downstream_refs": downstream_refs,
            "cores": cores,
            "partitions": partitions,
        }
        json_dict = json.dumps(ret_dict, cls=Enum2StrEncoder, indent=4)
        return json_dict


def parse_ltx_files(
    ltx_sources: List[Union[Path, str, TextIOBase]], source_names: Optional[List[str]] = None
) -> (Ltx, List[Union[str, TextIOBase]], List[str]):
    """
    One or more Ltx source(s) are read to produce one returned Ltx object.
    When multiple sources given, e.g. [<full Ltx file>, <partial ltx file>, <partial ltx file>]
    - RMs in later items override RMs for same re-configurable partitions found in sources
        to the left in the list.
    - A full Ltx file (or static Ltx file) will render obsolete all Ltx source items
        to the left in the list.

    ltx_sources and source_names, which are used for the resulting Ltx object, are returned.


    Args:
        ltx_sources: Ltx file path, or Ltx source text (io.StringIO or other TextIOBase type),
         or a list of such items.

        source_names: Name for the Ltx source, or a list of Ltx source names.
            Default is the Ltx_sources if file paths were given.
            Default for TextIOBase sources are "ltx_0", "ltx_1", ...

    Returns: A tuple with 3 members
        - Ltx object: which is the result of the combined Ltx sources.
        - List of Ltx_sources which contributed to the combined Ltx object.
        - List of source names, for the Ltx_sources which contributed to the combined Ltx object.

    """
    sources = ltx_sources if isinstance(ltx_sources, list) else [ltx_sources]
    if all([isinstance(source, (Path, str)) for source in sources]):
        # Clean up file paths.
        sources = [path.abspath(source) for source in sources]

    if not source_names:
        if all([isinstance(source, str) for source in sources]):
            source_names = sources
        else:
            source_names = ["ltx_" + str(idx) for idx in range(len(sources))]
    source_names = source_names if isinstance(source_names, list) else [source_names]
    if len(sources) != len(source_names):
        raise ValueError(
            'Function "ltx_files(ltx_sources, source_names)" needs to have '
            'equal number of items for arguments "ltx_sources" and "source_names".'
        )

    ltxs = []
    for ltx_src, ltx_name in zip(sources, source_names):
        ltx = Ltx(ltx_name)
        if isinstance(ltx_src, str):
            ltx.parse_file(ltx_src)
        else:
            ltx.parse_ltx(ltx_src)
        ltxs.append(ltx)

    if len(ltxs) == 1:
        # Only one source Ltx. Nothing to combine.
        return ltxs[0], list(sources), list(source_names)

    # Find last "Full Ltx" in the list. Assume first item is a "Full top-level Ltx"
    for idx in range(len(ltxs) - 1, 0, -1):
        if ltxs[idx].is_top_level():
            # Remove ltx files which do not matter, due to full re-programming of the device.
            ltxs = ltxs[idx:]
            sources = sources[idx:]
            source_names = source_names[idx:]
            break

    ltx: Ltx = ltxs[0]
    if len(ltxs) > 1:
        for partial_ltx in ltxs[1:]:
            ltx.replace_partition_cores(partial_ltx)
        ltx.post_process()

    names_of_sources_used = ltx.get_source_names()
    used_sources, used_source_names = zip(
        *[item for item in zip(sources, source_names) if item[1] in names_of_sources_used]
    )

    return ltx, list(used_sources), list(used_source_names)



# Generated from ILATsm.g4 by ANTLR 4.13.1
from antlr4 import *

if "." in __name__:
    from .ILATsmParser import ILATsmParser
else:
    from ILATsmParser import ILATsmParser

# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# This class defines a complete generic visitor for a parse tree produced by ILATsmParser.


class ILATsmVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by ILATsmParser#ila_tsm.
    def visitIla_tsm(self, ctx: ILATsmParser.Ila_tsmContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#state_if.
    def visitState_if(self, ctx: ILATsmParser.State_ifContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#state_no_if.
    def visitState_no_if(self, ctx: ILATsmParser.State_no_ifContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#if_condition.
    def visitIf_condition(self, ctx: ILATsmParser.If_conditionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#actions.
    def visitActions(self, ctx: ILATsmParser.ActionsContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#full_condition.
    def visitFull_condition(self, ctx: ILATsmParser.Full_conditionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#condition_counter_match.
    def visitCondition_counter_match(self, ctx: ILATsmParser.Condition_counter_matchContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#condition_paren.
    def visitCondition_paren(self, ctx: ILATsmParser.Condition_parenContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#condition_logic_op.
    def visitCondition_logic_op(self, ctx: ILATsmParser.Condition_logic_opContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#condition_probe_match.
    def visitCondition_probe_match(self, ctx: ILATsmParser.Condition_probe_matchContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#probe_match.
    def visitProbe_match(self, ctx: ILATsmParser.Probe_matchContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#counter_match.
    def visitCounter_match(self, ctx: ILATsmParser.Counter_matchContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_clear_flag.
    def visitAction_clear_flag(self, ctx: ILATsmParser.Action_clear_flagContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_set_flag.
    def visitAction_set_flag(self, ctx: ILATsmParser.Action_set_flagContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_inc_counter.
    def visitAction_inc_counter(self, ctx: ILATsmParser.Action_inc_counterContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_reset_counter.
    def visitAction_reset_counter(self, ctx: ILATsmParser.Action_reset_counterContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_goto.
    def visitAction_goto(self, ctx: ILATsmParser.Action_gotoContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#action_trigger.
    def visitAction_trigger(self, ctx: ILATsmParser.Action_triggerContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#probe_val.
    def visitProbe_val(self, ctx: ILATsmParser.Probe_valContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#enum_val.
    def visitEnum_val(self, ctx: ILATsmParser.Enum_valContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#enum_string_val.
    def visitEnum_string_val(self, ctx: ILATsmParser.Enum_string_valContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#int_val.
    def visitInt_val(self, ctx: ILATsmParser.Int_valContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#hex_val.
    def visitHex_val(self, ctx: ILATsmParser.Hex_valContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ILATsmParser#bin_val.
    def visitBin_val(self, ctx: ILATsmParser.Bin_valContext):
        return self.visitChildren(ctx)


del ILATsmParser



# *****************************************************************************
# * Copyright(C) 2020-2022 Xilinx, Inc.
# * Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Xilinx
# *****************************************************************************
try:
    import pytcf

    def to_xargs(args):
        return args


    def from_xargs(args):
        return args

except:
    import re
    import base64

    DEFAULT_ARG_VERSION = "arg_3"

    data_count_re = re.compile(r'\((\d+)\)')


    class JsonDataEncoder(object):
        def __init__(self, data):
            self.data = data

        def __json__(self):
            return "({})".format(len(self.data))

        def __str__(self):
            return self.__json__()

        def __repr__(self):
            return self.__json__()

    def preprocess_list(items):
        l = []
        data = bytearray()
        for v in items:
            t = type(v)
            if t == dict:
                v, d = preprocess_props(v)
                data += d
            elif t == bytearray or t == bytes:
                data += v
                v = JsonDataEncoder(v)
            l.append(v)
        return l, data


    def preprocess_props(props):
        out_props = {}
        data = bytearray()
        for k, v in props.items():
            t = type(v)
            if t == bytearray or t == bytes:
                data += v
                v = JsonDataEncoder(v)
            elif t == dict:
                v, d = preprocess_props(v)
                data += d
            elif t == list:
                v, d = preprocess_list(v)
                data += d
            out_props[k] = v
        return out_props, data


    def args_from_props(**props):
        args, data = preprocess_props(props)
        return DEFAULT_ARG_VERSION, args, data


    def to_xargs(args):
        xargs = []
        if type(args) == dict:
            props, data = preprocess_props(args)
            xargs.append(props)
            if len(data) > 0:
                xargs.append(data)
        else:
            for arg in args:
                if type(arg) == dict:
                    props, data = preprocess_props(arg)
                    xargs.append(props)
                    if len(data) > 0:
                        xargs.append(data)
                elif isinstance(arg, list) or isinstance(arg, tuple):
                    arg, data = preprocess_list(arg)
                    xargs.append(arg)
                    if len(data) > 0:
                        xargs.append(data)
                else:
                    xargs.append(arg)
        return xargs


    class ArgDataParser(object):
        def __init__(self, data):
            self.data = data
            self.pos = 0

        def __call__(self, size):
            end = self.pos + size
            if not isinstance(self.data, bytearray) and not isinstance(self.data, bytes):
                self.data = base64.b64decode(self.data)
            chunk = self.data[self.pos: end]
            self.pos = end
            return chunk


    def _dict_from_args(args, parser):
        return {k: _handle_arg_value(v, parser) for k, v in args.items()}


    def _handle_arg_value(arg_val, parser):
        t = type(arg_val)
        if t == dict:
            arg_val = _dict_from_args(arg_val, parser)
        elif t == list:
            arg_val = [_handle_arg_value(i, parser) for i in arg_val]
        elif t == str and parser:
            m = data_count_re.match(arg_val)
            if m:
                arg_val = parser(int(m.groups()[0]))
        return arg_val


    def props_from_args(args):
        props = from_xargs(args)

        if props and len(props) == 1:
            props = props[0]

        return props


    def from_xargs(args):
        props = []
        parser = None

        for i in range(len(args)):
            if args[i] == DEFAULT_ARG_VERSION:
                continue
            elif parser:
                parser = None
                continue
            elif i + 1 < len(args) and \
                    (type(args[i + 1]) == bytearray or type(args[i+1]) == str):
                parser = ArgDataParser(args[i + 1])
                props.append(_handle_arg_value(args[i], parser))
                if parser.pos == 0:  # parser unused so don't skip
                    parser = None
            # elif type(args[i]) == dict:
            #     if i+1 < len(args) and type(args[i+1]) == bytearray:
            #         parser = ArgDataParser(args[i+1])
            #     props.append(_dict_from_args(args[i], parser))
            else:
                props.append(args[i])
        return props



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
from chipscopy import dm
from chipscopy.client.core_property_client import CorePropertyClient
from chipscopy.tcf.services import DoneHWCommand
from chipscopy.proxies.HBMMCProxy import SERVICE_NAME


class HBMMCClient(CorePropertyClient):
    @staticmethod
    def is_compatible(node: dm.Node) -> bool:
        return node.type and node.type == "hbm_mc"

    def get_service_proxy(self):
        return self.manager.channel.getRemoteService(SERVICE_NAME)

    #
    #  General methods
    #
    def initialize(self, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        return self.add_pending(service.initialize(self.ctx, done_cb))

    def write32(self, write_addr: int, data: int, domain_index: int, done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.write32(self.ctx, write_addr, data, domain_index, done_cb)
        return self.add_pending(token)

    def read32(
        self, read_addr: int, read_word_count: int, domain_index: int, done: DoneHWCommand = None
    ):
        service, done_cb = self.make_done(done)
        token = service.read32(self.ctx, read_addr, read_word_count, domain_index, done_cb)
        return self.add_pending(token)

    #
    #  Service specific user interfaces
    #
    def refresh_property(self, names: List[str], done: DoneHWCommand = None):
        service, done_cb = self.make_done(done)
        token = service.refresh_property(self.ctx, names, done_cb)
        return self.add_pending(token)



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Dict, Union, Any

from chipscopy.api.ibert.aliases import CHILDREN, RX_KEY, TX_KEY, TYPE, PLL_SOURCE, PLL_KEY
from chipscopy.api.ibert.rx import RX
from chipscopy.api.ibert.serial_object_base import SerialObjectBase
from chipscopy.api.ibert.tx import TX
from chipscopy.api.ibert.pll import PLL
from more_itertools import one
from typing_extensions import final

if TYPE_CHECKING:  # pragma: no cover
    from chipscopy.api.ibert.gt_group import GTGroup  # noqa


@final
class GT(SerialObjectBase["GTGroup", Union[TX, RX]]):
    def __init__(self, gt_info: Dict[str, Any], parent, tcf_node):
        SerialObjectBase.__init__(self, gt_info, parent, tcf_node)

        # This is used by the filter_by method in QueryList
        self.filter_by = {"name": self.name, "type": self.type, "handle": self.handle}

        self.pll: list[PLL] = []
        "PLL driving this GT"

        self.rx: RX = None
        """RX of the GT"""

        self.tx: TX = None
        """TX of the GT"""

    # def __repr__(self) -> str:
    #     return self.name

    def reset(self):
        """
        Reset the TX and RX in the GT
        """
        self.setup()

        if self.rx:
            self.rx.reset()
        if self.tx:
            self.tx.reset()

    def update_pll(self):
        """Updates the reference to the pll attribute based on the appropriate property value"""
        if not self.setup_done:
            raise RuntimeError("Call this only after setup is complete for the GT")

        if PLL_SOURCE not in self._property_for_alias:
            return
        _, pll_names_for_this_gt = self._property.refresh(
            self._property_for_alias[PLL_SOURCE]
        ).popitem()
        if pll_names_for_this_gt:
            for pll in self.parent.plls:
                if pll.name == pll_names_for_this_gt:
                    self.pll.append(pll)
        else:
            print("No PLL defined for this GT")

    def setup(self):
        if self.setup_done:
            return

        obj_info = self.core_tcf_node.get_obj_info(self.handle)

        self._build_aliases(obj_info)

        if not obj_info.get(CHILDREN):
            return

        # Build the child objects
        for child_name, child_obj_info in obj_info[CHILDREN].items():
            obj: Union[TX, RX]
            if child_obj_info[TYPE] == RX_KEY:
                obj = RX(child_obj_info, self, self.core_tcf_node)
            elif child_obj_info[TYPE] == TX_KEY:
                obj = TX(child_obj_info, self, self.core_tcf_node)
            elif child_obj_info[TYPE].startswith(PLL_KEY):
                obj = PLL(child_obj_info, self, self.core_tcf_node)
            else:
                continue

            obj.setup()
            self._children.append(obj)

        # We always expect a GT to have one valid RX and one valid TX, no more, no less.
        self.rx = one(
            self._children.filter_by(type=RX_KEY),
            too_short=RuntimeError(f"No RX associated with GT {self.handle}!"),
            too_long=RuntimeError(f"More than one RX found for GT {self.handle}!"),
        )

        self.tx = one(
            self._children.filter_by(type=TX_KEY),
            too_short=RuntimeError(f"No TX associated with GT {self.handle}!"),
            too_long=RuntimeError(f"More than one TX found for GT {self.handle}!"),
        )

        self.setup_done = True



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
from io import TextIOBase, StringIO
from typing import List, Dict, Union, Optional
from antlr4 import CommonTokenStream, FileStream, InputStream

from chipscopy.api.ila.tsm.ILATsmLexer import ILATsmLexer
from chipscopy.api.ila.tsm.ILATsmParser import ILATsmParser
from chipscopy.api.ila.tsm.ILATsmVisitor import ILATsmVisitor
from chipscopy.api.ila.tsm.ila_tsm_data import ILATsmData, ILATsmErrorListener, MAX_STATE_COUNT
from chipscopy.api.ila.tsm.ila_tsm_checker import ILATsmCheckAndBuild
from chipscopy.api.ila import ILAPort, ILAProbe
from chipscopy.api.ila.tsm.ila_tsm_mapper import map_tsm_to_props


class ILATsmReader:
    def __init__(
        self,
        input_text: Union[TextIOBase, str],
        ports: List[ILAPort],
        probes: Dict[str, ILAProbe],
        probe_enum_def: Dict[str, enum.EnumMeta],
        counter_bit_widths: List[int],
        has_basic_capture_control: bool = False,
    ):
        self._mapper = None
        self._input = input_text
        self._input_stream: InputStream = None
        self._ports: List[ILAPort] = ports
        self._probes: Dict[str, ILAProbe] = probes
        self._probe_enum_defs: Dict[str, enum.EnumMeta] = probe_enum_def
        self._counter_bit_widths: List[int] = counter_bit_widths
        self._error_string: StringIO = StringIO()
        self._error_listener: ILATsmErrorListener = ILATsmErrorListener(self._error_string)
        self._has_basic_capture_control = has_basic_capture_control
        self._state_names: Dict[int, str] = {}

    def get_error_count(self):
        return self._error_listener.get_error_count()

    def get_error_str(self):
        return self._error_string.getvalue()

    def get_state_names(self) -> Dict[int, str]:
        return self._state_names if self.get_error_count() == 0 else {}

    def parse(self, compile_only: bool = False) -> (int, str, Optional[dict]):
        """

        Args:
            compile_only (bool): If True, the ILA core register values are not generated.

        Returns: (error_count, error_message, register settings).
            register setting will be None if error_count > 0 or compile_only is True.
        """

        tsm_data = None
        state_names = []
        if isinstance(self._input, TextIOBase):
            self._input_stream = InputStream(self._input.read())
        else:
            self._input_stream = FileStream(self._input)

        # Setup lexer
        lexer = ILATsmLexer(self._input_stream)
        lexer.removeErrorListeners()
        lexer.addErrorListener(self._error_listener)

        # Parsing
        token_stream = CommonTokenStream(lexer)
        parser = ILATsmParser(token_stream)
        parser.removeErrorListeners()
        parser.addErrorListener(self._error_listener)
        tree = parser.ila_tsm()
        if self.get_error_count() == 0:
            tsm_data = ILATsmData("", self._ports, self._probes, self._counter_bit_widths)

            # Scan for state names
            state_visitor = ILATsmStateVisitor(self._error_listener)
            state_visitor.visit(tree)
            state_names = state_visitor.get_states()
            if len(state_names) > MAX_STATE_COUNT:
                self._error_listener.report_error(
                    0,
                    0,
                    f"Maximum {MAX_STATE_COUNT} states are allowed. "
                    f"The Trigger State Machine has {len(state_names)} states.",
                )
            self._state_names = {idx: name for idx, name in enumerate(state_names)}

        # Semantic checking and build back-end data structure.
        if self.get_error_count() == 0:
            checker = ILATsmCheckAndBuild(
                tsm_data,
                self._probe_enum_defs,
                state_names,
                self._counter_bit_widths,
                self._error_listener,
            )
            checker.visit(tree)

        props = None
        if not compile_only and self.get_error_count() == 0:
            props = map_tsm_to_props(
                tsm_data, self._has_basic_capture_control, self._error_listener
            )
        return self.get_error_count(), self.get_error_str(), props


class ILATsmStateVisitor(ILATsmVisitor):
    """
    Find state names.
    """

    def __init__(self, error_listener: ILATsmErrorListener):
        self._states = []
        self._error_listener = error_listener

    def get_states(self):
        return self._states

    def _add_state(self, ctx: Union[ILATsmParser.State_ifContext, ILATsmParser.State_no_ifContext]):
        state_name = ctx.IDENTIFIER().getText()
        if state_name in self._states:
            self._error_listener.report_error_ctx(
                ctx, f'Two states have the same name: "{state_name}". State names must be unique.'
            )
        else:
            self._states.append(state_name)

    # Visit a parse tre0e produced by ILATsmParser#state_if.
    def visitState_if(self, ctx: ILATsmParser.State_ifContext):
        self._add_state(ctx)
        return None

    # Visit a parse tree produced by ILATsmParser#state_no_if.
    def visitState_no_if(self, ctx: ILATsmParser.State_no_ifContext):
        self._add_state(ctx)
        return None



# Copyright (C) 2022, Xilinx, Inc.
# Copyright (C) 2022-2023, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""device_utils.py -

Collection of utility functions

"""
from typing import Optional

from chipscopy.client import ServerInfo
from chipscopy.client.jtagdevice import JtagDevice, JtagRegister
from chipscopy.client.view_info import ViewInfo
from chipscopy.dm import Node


def get_node_hier_name(view: ViewInfo, node: Node) -> str:
    name = node.props["Name"]
    while node.parent_ctx:
        prev_node = node
        node = view.get_node(node.parent_ctx)
        if node == prev_node:
            break
        parent_name = node.props["Name"]
        name = f"{parent_name}/{name}"
    return name


def copy_node_props(node: Node, props: dict, *, force_update=True) -> dict:
    transformed_props = {}
    for k, v in props.items():
        if type(v) == int or type(v) == str or type(v) == bool:
            transformed_props[str(k)] = v
        elif type(v) == JtagRegister:
            try:
                if force_update:
                    node.update_regs(reg_names=(f"{k}",), force=True, done=None)
                list_val = [
                    int.from_bytes(reg_data, byteorder="little", signed=False)
                    for reg_data in v.data
                ]
                transformed_props[str(k)] = list_val
            except Exception:  # noqa
                pass
        elif type(v) == dict:
            transformed_props[str(k)] = copy_node_props(node, v, force_update=force_update)
        elif type(v) == bytearray:
            int_val = int.from_bytes(v, byteorder="little", signed=False)
            transformed_props[str(k)] = int_val
        else:
            transformed_props[str(k)] = str(v)
    return transformed_props


def _create_jtag_node_dict(
    server_name: str,
    view_name: str,
    jtag_cable_ctx: str,
    jtag_index: int,
    view: ViewInfo,
    node: Node,
) -> dict:
    children = list(view.get_children(node))
    jtag_node_dict = {
        "server": server_name,
        "view": view_name,
        "ctx": node.ctx,
        "dna": get_node_dna(node),
        "family": node.props.get("arch_name", None),
        "name": node.props["Name"],
        "hier_name": get_node_hier_name(view, node),
        "jtag_cable_ctx": jtag_cable_ctx,
        "jtag_index": jtag_index,
        "props": copy_node_props(node, node.props),
        "parent_ctx": node["ParentID"],
        "children": {
            child.ctx: _create_jtag_node_dict(
                server_name, view_name, jtag_cable_ctx, jtag_index, view, child
            )
            for child in children
        },
    }
    return jtag_node_dict


def get_jtag_view_dict(hw_server: ServerInfo) -> dict:
    jtag_dict = {}
    server_name = "hw_server"
    view_name = "jtag"
    view = hw_server.get_view(view_name)
    for jtag_cable in view.get_children():
        device_dict = {}
        jtag_cable_ctx = jtag_cable.ctx
        device_nodes_to_process = list(view.get_children(jtag_cable))
        jtag_index = 0
        for node in device_nodes_to_process:
            device_dict[node.ctx] = _create_jtag_node_dict(
                server_name, view_name, jtag_cable_ctx, jtag_index, view, node
            )
        jtag_dict[jtag_cable_ctx] = {
            "name": jtag_cable.props["Name"],
            "ctx": jtag_cable_ctx,
            "devices": device_dict,
            "props": copy_node_props(jtag_cable, jtag_cable.props),
        }
    return jtag_dict


def _create_basic_node_dict(server_name: str, view_name: str, view: ViewInfo, node: Node) -> dict:
    children = list(view.get_children(node))
    debugcore_node_dict = {
        "server": server_name,
        "view": view_name,
        "ctx": node.ctx,
        "dna": get_node_dna(node),
        "name": node.props["Name"],
        "hier_name": get_node_hier_name(view, node),
        "props": copy_node_props(node, node.props),
        "children": {
            child.ctx: _create_basic_node_dict(server_name, view_name, view, child)
            for child in children
        },
    }
    return debugcore_node_dict


def _get_basic_view_dict(server: ServerInfo, server_name, view_name) -> dict:
    device_dict = {}
    view = server.get_view(view_name)
    for node in view.get_children():
        device_dict[node.ctx] = _create_basic_node_dict(server_name, view_name, view, node)
    return device_dict


def get_memory_view_dict(hw_server: ServerInfo) -> dict:
    return _get_basic_view_dict(hw_server, "hw_server", "memory")


def get_debugcore_view_dict(hw_server: ServerInfo) -> dict:
    return _get_basic_view_dict(hw_server, "hw_server", "debugcore")


def get_chipscope_view_dict(cs_server: ServerInfo) -> dict:
    return _get_basic_view_dict(cs_server, "cs_server", "chipscope")


def get_node_dna(node: Node) -> Optional[int]:
    """Returns: 128-bit device dna value if available for the node, None otherwise."""
    dna_128 = None
    if node.props.get("node_cls") is JtagDevice:
        # jtag devices store dna as a bytestream
        if node.props.get("regs", None):
            if node.props["regs"].get("dna", None):
                # Below refreshes the dna data and makes it available
                # node.status('dna')
                try:
                    node.update_regs(reg_names=("dna",), force=True, done=None)
                    jtag_register = node.props["regs"]["dna"]
                    bytearray_data = jtag_register.data[0]
                    dna_128 = int.from_bytes(bytearray_data, byteorder="little", signed=False)
                except Exception:  # noqa
                    dna_128 = None
    else:
        # Other contexts store dna as a 4-tuple of 32-bit ints
        dna_4_tuple = node.props.get("DeviceDNA", None)
        if dna_4_tuple and len(dna_4_tuple) == 4:
            dna_128 = (
                dna_4_tuple[3] << 96 | dna_4_tuple[2] << 64 | dna_4_tuple[1] << 32 | dna_4_tuple[0]
            )
    return dna_128



# Copyright (C) 2021-2022, Xilinx, Inc.
# Copyright (C) 2022-2024, Advanced Micro Devices, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytcf

SERVICE_NAME = "Locator"

_waiting_for_disconnect = {}


def get_channels(token, channel):
    args = pytcf.read_command_args(channel)
    pytcf.write_response_args(token, channel, None, ([c.url for c in pytcf.get_channels()],))


def connect_peer(token, channel):
    args = pytcf.read_command_args(channel)

    if len(args) < 1:
        pytcf.write_response_args(token, channel, pytcf.ERR_INV_CONTEXT)
        return

    attr = args[0]
    url = ""
    if type(attr) == dict:
        url = attr.get("ID")
    elif type(attr) == str:
        url = attr

    if not url:
        pytcf.write_response_args(token, channel, pytcf.ERR_INV_CONTEXT)
        return

    if url[:3] != "TCP":
        url = "TCP:" + url
    target = pytcf.get_channel_from_url(url)
    if target:
        # target.add_linked_channel(channel)
        pytcf.write_response_args(token, channel, None, (target.url,))
        return

    def connect_done(req):
        err = None
        target = pytcf.get_channel_from_url(req.url)
        if target:
            # target.add_linked_channel(channel)
            pass
        else:
            err = pytcf.errno
        pytcf.write_response_args(token, channel, err, (target.url,))

    pytcf.connect_tcf(url, done=connect_done)


def disconnect_peer(token, channel):
    args = pytcf.read_command_args(channel)

    if len(args) < 1:
        pytcf.write_response_args(token, channel, pytcf.ERR_INV_CONTEXT)
        return

    attr = args[0]
    url = ""
    if type(attr) == dict:
        url = attr.get("ID")
    elif type(attr) == str:
        url = attr

    target = pytcf.get_channel_from_url(url)
    if not target:
        pytcf.write_response_args(token, channel, pytcf.ERR_INV_CONTEXT)
        return

    def peer_disconnected():
        pytcf.write_response_args(token, channel, None, (True,))

    _waiting_for_disconnect[target] = peer_disconnected
    pytcf.disconnect_channel(target)


def channel_closed(channel):
    channels = list(pytcf.get_channels())
    for c in channels:
        if channel in c.linked_channels:
            c.linked_channels.remove(channel)
            if not c.linked_channels:
                pytcf.disconnect_channel(c)

    f = _waiting_for_disconnect.get(channel)
    if f:
        f()


def ini_service(si):
    pytcf.add_command_handler(si, SERVICE_NAME, "getChannels", get_channels)
    pytcf.add_command_handler(si, SERVICE_NAME, "connectPeer", connect_peer)
    pytcf.add_command_handler(si, SERVICE_NAME, "disconnectPeer", disconnect_peer)
    pytcf.add_channel_close_listener(channel_closed)


# *****************************************************************************
# * Modifications Copyright (C) 2021-2022, Xilinx, Inc., All rights reserved.
# * Modifications Copyright (C) 2022-2024, Advanced Micro Devices, Inc., All rights reserved..
# *
# * Copyright (c) 2011, 2013 Wind River Systems, Inc. and others.
# * All rights reserved. This program and the accompanying materials
# * are made available under the terms of the Eclipse Public License 2.0
# * which accompanies this distribution, and is available at
# * https://www.eclipse.org/legal/epl-2.0/
# *
# * Contributors:
# *     Wind River Systems - initial API and implementation
# *     Xilinx
# *****************************************************************************

import time
import pytcf
from typing import ClassVar
from .. import Service
from ... import protocol


class CommandRequest(object):
    def __init__(self, _channel, token, data, node=None):
        self.channel = _channel
        self.token = token
        self.data = data
        if node is not None:
            self.node = node
        self.args = data

    def sendError(self, error, *result):
        if type(error) is list:
            error_props = {
                "Code": 0x20001,  # ERR_OTHER default used by hw_server
                "Time": int(time.perf_counter() * 1000),
                "Format": str(error),
            }
        else:
            error_props = {
                "Code": 0x20001,  # ERR_OTHER default used by hw_server
                "Time": int(time.perf_counter() * 1000),
                "Format": str(error),
                "Class": error.__class__.__name__,
                "Module": error.__class__.__module__
            }
        #self.channel.sendResult(self.token, (error_props, *result))
        pytcf.write_response_args(self.token, self.channel, error_props, None)

    def sendResult(self, *result):
        #self.channel.sendResult(self.token, (None, *result))
        pytcf.write_response_args(self.token, self.channel, None, result)

    def sendXicomResult(self, *results):
        #self.channel.sendResult(self.token, (None, *results))
        pytcf.write_response_args(self.token, self.channel, None, results)

    def sendXicomProgress(self, *progress):
        #self.channel.sendProgress(self.token, (None, *progress))
        pytcf.write_progress_args(self.token, self.channel, progress)


def send_xicom_event(service_name, event_name, args):
    protocol.sendEvent(service_name, event_name, args)


def _wrap_command(callback):
    def handle_command(token, channel):
        args = pytcf.read_command_args(channel)
        callback(CommandRequest(channel, token, args))
    return handle_command


def to_camel_case(name: str) -> str:
    m = name.split("_")
    return m[0] + "".join(i.title() for i in m[1:-1])


_local_services_added = set()
_service_interface = None


def init(si):
    global _service_interface
    _service_interface = si


def add_local_service(service: Service or ClassVar[Service]):
    convert_camel_case = True
    if hasattr(service, "convert_camel_case"):
        convert_camel_case = service.convert_camel_case
    if str(service) not in _local_services_added:
        _local_services_added.add(str(service))
        for c in [c for c in dir(service) if not c.startswith("__") and c.endswith("_cmd")]:
            pytcf.add_command_handler(
                _service_interface,
                service.getName(),
                to_camel_case(c) if convert_camel_case else c[:c.find("_cmd")],
                _wrap_command(getattr(service, c))
            )
        for c in [c for c in dir(service) if not c.startswith("__") and c.endswith("Cmd")]:
            pytcf.add_command_handler(
                _service_interface,
                service.getName(),
                c[:c.find("Cmd")],
                _wrap_command(getattr(service, c))
            )
